# Copyright (c) 2023 China Unicom Cloud Data Co.,Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import os
import secrets
import time

import netaddr
from neutron_lib.agent import l2_extension as l2_agent_extension
from neutron_lib import constants
from neutron_lib import exceptions as n_exc
from neutron_lib.plugins.ml2 import ovs_constants as p_const
from neutron_lib.plugins import utils as p_utils
from neutron_lib.utils import net as net_lib
from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging

from neutron._i18n import _
from neutron.agent.common import ip_lib
from neutron.agent.l2.extensions.metadata import host_metadata_proxy
from neutron.agent.l2.extensions.metadata import metadata_flows_process
from neutron.agent.linux import external_process
from neutron.api.rpc.callbacks import resources
from neutron.common import coordination

LOG = logging.getLogger(__name__)

DEFAULT_META_GATEWAY_MAC = "fa:16:ee:00:00:01"
METADATA_DEFAULT_MAC = 'fa:16:ee:ff:ff:ff'


class InvalidProviderCIDR(n_exc.NeutronException):
    message = _("Not enough Metadata IPs in /32 CIDR")


class NoMoreProviderRes(n_exc.NeutronException):
    message = _("No more %(res)s")


class FailedToInitMetadataPathExtension(n_exc.NeutronException):
    message = _("Could not initialize agent extension "
                "metadata path, error: %(msg)s")


class MetadataPathExtensionPortInfoAPI:

    def __init__(self, cache_api):
        self.cache_api = cache_api
        self.allocated_ips = netaddr.IPSet()
        self.allocated_macs = set()

    def get_port_fixed_ip(self, port):
        for ip in port.fixed_ips:
            ip_addr = netaddr.IPAddress(str(ip.ip_address))
            if ip_addr.version == constants.IP_VERSION_4:
                return str(ip.ip_address)

    def remove_allocated_ip(self, ip):
        self.allocated_ips.remove(ip)

    def remove_allocated_mac(self, mac):
        self.allocated_macs.remove(mac)

    def _get_one_ip(self):

        def generate_local_ip(cidr):
            network = netaddr.IPNetwork(cidr)
            if network.prefixlen == 32:
                raise InvalidProviderCIDR()
            # https://docs.python.org/3/library/secrets.html#module-secrets
            # secrets.randbelow(exclusive_upper_bound)
            # Return a random int in the range [0, exclusive_upper_bound).
            # Here we remove the first and last IPs here.
            index = secrets.randbelow(network.size - 1)
            return str(network[index + 1])

        for _i in range(1, 100):
            ip = generate_local_ip(cfg.CONF.METADATA.provider_cidr)
            if ip not in self.allocated_ips:
                return ip
        raise NoMoreProviderRes(res="provider IP addresses")

    def _get_one_mac(self):
        for _i in range(1, 1000):
            base_mac = cfg.CONF.METADATA.provider_base_mac
            mac = net_lib.get_random_mac(base_mac.split(':'))
            if mac not in self.allocated_macs:
                return mac
        raise NoMoreProviderRes(res="provider MAC addresses")

    def get_provider_ip_info(self, port_id,
                             provider_ip=None,
                             provider_mac=None):
        port_obj = self.cache_api.get_resource_by_id(
            resources.PORT, port_id)
        if not port_obj or not port_obj.device_id:
            return

        info = {"instance_id": port_obj.device_id,
                "project_id": port_obj.project_id}

        if (not provider_ip or netaddr.IPNetwork(provider_ip) not in
                netaddr.IPNetwork(cfg.CONF.METADATA.provider_cidr)):
            provider_ip = self._get_one_ip()
        self.allocated_ips.add(provider_ip)
        info["provider_ip"] = provider_ip

        if not provider_mac:
            provider_mac = self._get_one_mac()
        self.allocated_macs.add(provider_mac)
        info["provider_port_mac"] = provider_mac

        return info


class MetadataPathAgentExtension(
        l2_agent_extension.L2AgentExtension,
        metadata_flows_process.MetadataDataPathFlows):

    PORT_INFO_CACHE = {}
    META_DEV_NAME = "tap-meta"

    NETWORK_DHCP_PORTS = {}
    NETWORK_PORTS = collections.defaultdict(set)

    @lockutils.synchronized('networking-path-ofport-cache')
    def set_port_info_cache(self, port_id, port_info):
        self.PORT_INFO_CACHE[port_id] = port_info

    @lockutils.synchronized('networking-path-ofport-cache')
    def get_port_info_from_cache(self, port_id):
        return self.PORT_INFO_CACHE.pop(port_id, None)

    def consume_api(self, agent_api):
        if not all([agent_api.br_phys.get('meta'), agent_api.phys_ofports,
                    agent_api.bridge_mappings.get('meta')]):
            raise FailedToInitMetadataPathExtension(
                msg="The metadata bridge device may not exist.")

        self.agent_api = agent_api
        self.rcache_api = agent_api.plugin_rpc.remote_resource_cache

    def initialize(self, connection, driver_type):
        """Initialize agent extension."""
        self.ext_api = MetadataPathExtensionPortInfoAPI(self.rcache_api)
        self.int_br = self.agent_api.request_int_br()
        self.meta_br = self.agent_api.request_physical_br('meta')
        self.instance_infos = {}

        bridge = self.agent_api.bridge_mappings.get('meta')
        port_name = p_utils.get_interface_name(
            bridge, prefix=p_const.PEER_INTEGRATION_PREFIX)
        self.ofport_int_to_meta = self.int_br.get_port_ofport(port_name)
        self.ofport_meta_to_int = self.agent_api.phys_ofports['meta']

        if (not cfg.CONF.METADATA.nova_metadata_host or
                not cfg.CONF.METADATA.nova_metadata_port):
            LOG.warning("Nova metadata API related options are not set. "
                        "Host metadata haproxy will not start. "
                        "Please check the config option of "
                        "'nova_metadata_*' in [METADATA] section.")
            return
        self.process_monitor = external_process.ProcessMonitor(
            config=cfg.CONF,
            resource_type='MetadataPath')
        self.meta_daemon = host_metadata_proxy.HostMedataHAProxyDaemonMonitor(
            self.process_monitor,
            user=str(os.geteuid()),
            group=str(os.getegid()))

        self.provider_vlan_id = cfg.CONF.METADATA.provider_vlan_id
        self.provider_cidr = cfg.CONF.METADATA.provider_cidr

        self.provider_gateway_ip = str(netaddr.IPAddress(
            netaddr.IPNetwork(cfg.CONF.METADATA.provider_cidr).first + 1))

        self._create_internal_port()

        # In order to cover the situation that the VM has a link reachable
        # route to 169.254.169.254.
        self.install_arp_responder(
            bridge=self.int_br,
            ip=metadata_flows_process.METADATA_V4_IP,
            mac=METADATA_DEFAULT_MAC,
            table=p_const.TRANSIENT_TABLE)

        self.set_path_br(self.meta_br)
        self.metadata_ofport = self.meta_br.get_port_ofport(self.META_DEV_NAME)
        self.init_br_snat_metadata_path()

    def _set_port_vlan(self):
        ovsdb = self.meta_br.ovsdb
        with self.meta_br.ovsdb.transaction() as txn:
            # When adding the port's tag,
            # also clear port's vlan_mode and trunks,
            # which were set to make sure all packets are dropped.
            txn.add(ovsdb.db_set('Port', self.META_DEV_NAME,
                                 ('tag', self.provider_vlan_id)))
            txn.add(ovsdb.db_clear('Port', self.META_DEV_NAME, 'vlan_mode'))
            txn.add(ovsdb.db_clear('Port', self.META_DEV_NAME, 'trunks'))

    def _create_internal_port(self):
        attrs = [('type', 'internal'),
                 ('external_ids', {'iface-status': 'active',
                                   'attached-mac': DEFAULT_META_GATEWAY_MAC})]
        self.meta_br.replace_port(self.META_DEV_NAME, *attrs)

        ns_dev = ip_lib.IPDevice(self.META_DEV_NAME)

        for _i in range(9):
            try:
                ns_dev.link.set_address(DEFAULT_META_GATEWAY_MAC)
                break
            except RuntimeError as e:
                LOG.warning("Got error trying to set mac, retrying: %s", e)
                time.sleep(1)

        try:
            ns_dev.link.set_address(DEFAULT_META_GATEWAY_MAC)
        except RuntimeError as e:
            raise RuntimeError(
                _("Failed to set mac address "
                  "for dev %(dev)s, error: %(error)s") %
                  {'dev': self.META_DEV_NAME, 'error': e})

        cidr = "{}/{}".format(
            self.provider_gateway_ip,
            netaddr.IPNetwork(self.provider_cidr).prefixlen)
        ns_dev.addr.add(cidr)
        ns_dev.link.set_up()

        self.meta_br.set_value_to_other_config(
            self.META_DEV_NAME,
            "tag",
            self.provider_vlan_id)
        self._set_port_vlan()

    def _reload_host_metadata_proxy(self, force_reload=False):
        if (not cfg.CONF.METADATA.nova_metadata_host or
                not cfg.CONF.METADATA.nova_metadata_port):
            LOG.warning("Nova metadata API related options are not set. "
                        "Host metadata haproxy will not start.")
            return
        if not force_reload and not self.instance_infos:
            return
        # Haproxy does not suport 'kill -HUP' to reload config file,
        # so just kill it and then re-spawn.
        self.meta_daemon.disable()
        self.meta_daemon.config(list(self.instance_infos.values()))
        if self.instance_infos:
            self.meta_daemon.enable()

    def _get_port_info(self, port_detail):
        device_owner = port_detail['device_owner']
        if not device_owner.startswith(constants.DEVICE_OWNER_COMPUTE_PREFIX):
            return

        port = port_detail['vif_port']
        provider_ip = self.int_br.get_value_from_other_config(
            port.port_name, 'provider_ip')
        provider_mac = self.int_br.get_value_from_other_config(
            port.port_name, 'provider_mac')

        ins_info = self.ext_api.get_provider_ip_info(port_detail['port_id'],
                                                     provider_ip,
                                                     provider_mac)
        if not ins_info:
            LOG.info("Failed to get port %s instance provider IP info.",
                     port_detail['port_id'])
            return
        self.instance_infos[port_detail['port_id']] = ins_info
        if not provider_ip or provider_ip != ins_info['provider_ip']:
            self.int_br.set_value_to_other_config(
                port.port_name,
                'provider_ip',
                ins_info['provider_ip'])
        if not provider_mac:
            self.int_br.set_value_to_other_config(
                port.port_name,
                'provider_mac',
                ins_info['provider_port_mac'])

        vlan = self.int_br.get_value_from_other_config(
            port.port_name, 'tag', int)

        port_info = {"port_id": port_detail['port_id'],
                     "device_owner": device_owner,
                     "port_name": port.port_name,
                     "vlan": vlan,
                     "mac_address": port_detail["mac_address"],
                     "fixed_ips": port_detail["fixed_ips"],
                     "ofport": port.ofport,
                     "network_id": port_detail['network_id']}

        LOG.debug("Metadata path got the port information: %s ",
                  port_info)
        return port_info

    def handle_port(self, context, port_detail):
        try:
            port_info = self._get_port_info(port_detail)
            if not port_info:
                return
            self.set_port_info_cache(port_detail['port_id'], port_info)
        except Exception as err:
            LOG.info("Failed to get or set port %s info, error: %s",
                     port_detail['port_id'], err)
        else:
            self.process_install_metadata_path_flows(port_info)
            self.install_dhcp_ports_arp_responder(port_info)
            self._reload_host_metadata_proxy()

    def _get_fixed_ip(self, port_info):
        for ip in port_info['fixed_ips']:
            ip_addr = netaddr.IPAddress(ip['ip_address'])
            if ip_addr.version == constants.IP_VERSION_4:
                return ip['ip_address']

    def delete_port(self, context, port_detail):
        ins_info = self.instance_infos.pop(port_detail['port_id'], None)
        self._reload_host_metadata_proxy(force_reload=True)
        port_info = self.get_port_info_from_cache(port_detail['port_id'])
        if not port_info:
            LOG.info("No port_info cache found for %s, "
                     "skipping remove networking path related flows.",
                     port_detail['port_id'])
            return

        self.remove_port_metadata_direct_flow(port_info["ofport"],
                                              port_info["vlan"],
                                              port_info['mac_address'],
                                              self.ofport_int_to_meta)
        self.remove_dhcp_ports_arp_responder(port_info)

        if not ins_info:
            return
        self.ext_api.remove_allocated_ip(ins_info['provider_ip'])
        self.ext_api.remove_allocated_mac(ins_info['provider_port_mac'])

        port_fixed_ip = self._get_fixed_ip(port_info)
        if not port_fixed_ip:
            return

        self.remove_port_metadata_path_nat_and_arp_flow(
            port_info['vlan'], port_info['mac_address'], port_fixed_ip,
            self.provider_vlan_id, ins_info["provider_ip"])

    def init_br_snat_metadata_path(self):
        self.metadata_host_info = {
            "gateway_ip": self.provider_gateway_ip,
            "provider_ip": self.provider_gateway_ip,
            "mac_address": DEFAULT_META_GATEWAY_MAC,
            "service_protocol_port": cfg.CONF.METADATA.host_proxy_listen_port}
        self.metadata_path_defaults(
            pvid=self.provider_vlan_id,
            to_int_ofport=self.ofport_meta_to_int,
            metadata_ofport=self.metadata_ofport,
            metadata_host_info=self.metadata_host_info)

    def process_install_metadata_path_flows(self, port_info):
        if not self.meta_br:
            return

        try:
            self.install_port_metadata_path_direct_flows(port_info)
        except Exception as err:
            LOG.debug("Failed to install networking path direct flows "
                      "for port %s, error: %s",
                      port_info['port_id'], err)
        try:
            self.install_port_metadata_path_nat_and_arp_flows(port_info)
        except Exception as err:
            LOG.debug("Failed to install networking path nat and "
                      "ARP responder flows for port %s, error: %s",
                      port_info['port_id'], err)

    def install_port_metadata_path_nat_and_arp_flows(
            self, port_info):
        phys_port = self.agent_api.phys_ofports['meta']
        info = self.instance_infos.get(port_info['port_id'])
        if not info:
            return
        provider_ip_addr = info["provider_ip"]
        provider_port_mac = info["provider_port_mac"]

        port_fixed_ip = self._get_fixed_ip(port_info)
        if not port_fixed_ip:
            return

        self.add_flow_snat_br_meta(
            port_info["vlan"], port_info['mac_address'], port_fixed_ip,
            provider_port_mac, provider_ip_addr)
        self.install_arp_responder(
            ip=provider_ip_addr,
            mac=provider_port_mac)
        self.add_flow_ingress_dnat_direct_to_int_br(
            port_info["vlan"], self.provider_vlan_id, provider_ip_addr,
            port_info['mac_address'], port_fixed_ip,
            phys_port,
            self.metadata_ofport)

    def install_port_metadata_path_direct_flows(self, port_info):
        self.add_flow_int_br_egress_direct(
            port_info["ofport"], port_info["vlan"],
            self.ofport_int_to_meta)
        self.add_flow_int_br_ingress_output(
            self.ofport_int_to_meta, port_info["vlan"],
            port_info['mac_address'], port_info["ofport"])

    @coordination.synchronized('meta-net-{port_info[network_id]}')
    def install_dhcp_ports_arp_responder(self, port_info):
        filters = {'network_id': (port_info['network_id'], ),
                   'device_owner': (constants.DEVICE_OWNER_DHCP, )}
        dhcp_ports = self.rcache_api.get_resources(resources.PORT, filters)
        for p in dhcp_ports:
            for ip in p['fixed_ips']:
                ip_addr = netaddr.IPNetwork(str(ip.ip_address))
                if ip_addr.version != constants.IP_VERSION_4:
                    continue
                self.install_arp_responder(
                    bridge=self.int_br,
                    ip=str(ip.ip_address),
                    mac=str(p.mac_address),
                    table=p_const.TRANSIENT_TABLE,
                    in_port=port_info["ofport"])

        self.NETWORK_DHCP_PORTS[port_info['network_id']] = dhcp_ports
        self.NETWORK_PORTS[port_info['network_id']].add(port_info['port_id'])

    @coordination.synchronized('meta-net-{port_info[network_id]}')
    def remove_dhcp_ports_arp_responder(self, port_info):
        if self.NETWORK_PORTS[port_info['network_id']]:
            self.NETWORK_PORTS[port_info['network_id']].remove(
                port_info['port_id'])

        # No ports from this network, remove the DHCP ports ARP responder.
        if not self.NETWORK_PORTS[port_info['network_id']]:
            dhcp_ports = self.NETWORK_DHCP_PORTS.pop(
                port_info['network_id'], [])
            for p in dhcp_ports:
                for ip in p['fixed_ips']:
                    ip_addr = netaddr.IPNetwork(str(ip.ip_address))
                    if ip_addr.version != constants.IP_VERSION_4:
                        continue
                    self.delete_arp_responder(
                        bridge=self.int_br,
                        ip=str(ip.ip_address),
                        table=p_const.ARP_SPOOF_TABLE,
                        in_port=port_info["ofport"])
            self.NETWORK_PORTS.pop(port_info['network_id'], [])



# Copyright (c) 2014 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

from neutron_lib import constants
from oslo_config import cfg
from oslo_log import log as logging

from neutron._i18n import _
from neutron.agent import dhcp_agent
from neutron.cmd.sanity import checks
from neutron.common import config
from neutron.conf.agent import common as agent_config
from neutron.conf.agent import securitygroups_rpc
from neutron.conf import common as common_config
from neutron.conf.db import l3_hamode_db
from neutron.conf.plugins.ml2 import config as ml2_conf
from neutron.conf.plugins.ml2.drivers.mech_sriov import agent_common as \
    sriov_conf
from neutron.conf.plugins.ml2.drivers import ovs_conf


LOG = logging.getLogger(__name__)


def setup_conf():
    config.register_common_config_options()
    ovs_conf.register_ovs_agent_opts(cfg.CONF)
    sriov_conf.register_agent_sriov_nic_opts(cfg.CONF)
    ml2_conf.register_ml2_plugin_opts(cfg.CONF)
    securitygroups_rpc.register_securitygroups_opts(cfg.CONF)
    dhcp_agent.register_options(cfg.CONF)
    l3_hamode_db.register_db_l3_hamode_opts(cfg.CONF)
    common_config.register_core_common_config_opts(cfg.CONF)


class BoolOptCallback(cfg.BoolOpt):
    def __init__(self, name, callback, **kwargs):
        if 'default' not in kwargs:
            kwargs['default'] = False
        self.callback = callback
        super().__init__(name, **kwargs)


def check_ovs_vxlan():
    result = checks.ovs_vxlan_supported()
    if not result:
        LOG.error('Check for Open vSwitch VXLAN support failed. '
                  'Please ensure that the version of openvswitch '
                  'being used has VXLAN support.')
    return result


def check_ovs_geneve():
    result = checks.ovs_geneve_supported()
    if not result:
        LOG.error('Check for Open vSwitch Geneve support failed. '
                  'Please ensure that the version of openvswitch '
                  'and kernel being used has Geneve support.')
    return result


def check_iproute2_vxlan():
    result = checks.iproute2_vxlan_supported()
    if not result:
        LOG.error('Check for iproute2 VXLAN support failed. Please ensure '
                  'that the iproute2 has VXLAN support.')
    return result


def check_ovs_patch():
    result = checks.patch_supported()
    if not result:
        LOG.error('Check for Open vSwitch patch port support failed. '
                  'Please ensure that the version of openvswitch '
                  'being used has patch port support or disable features '
                  'requiring patch ports (gre/vxlan, etc.).')
    return result


def check_read_netns():
    required = checks.netns_read_requires_helper()
    if not required and cfg.CONF.AGENT.use_helper_for_ns_read:
        LOG.warning("The user that is executing neutron can read the "
                    "namespaces without using the root_helper. Disable "
                    "the use_helper_for_ns_read option to avoid a "
                    "performance impact.")
        # Don't fail because nothing is actually broken. Just not optimal.
        result = True
    elif required and not cfg.CONF.AGENT.use_helper_for_ns_read:
        LOG.error("The user that is executing neutron does not have "
                  "permissions to read the namespaces. Enable the "
                  "use_helper_for_ns_read configuration option.")
        result = False
    else:
        # everything is configured appropriately
        result = True
    return result


# NOTE(ihrachyshka): since the minimal version is currently capped due to
# missing hwaddr matching in dnsmasq < 2.67, a better version of the check
# would actually start dnsmasq server and issue a DHCP request using a IPv6
# DHCP client.
def check_dnsmasq_version():
    result = checks.dnsmasq_version_supported()
    if not result:
        LOG.error('The installed version of dnsmasq is too old. '
                  'Please update to at least version %s.',
                  checks.get_minimal_dnsmasq_version_supported())
    return result


def check_ovs_qos_direct_ports_supported():
    result = checks.ovs_qos_direct_port_supported()
    if not result:
        LOG.error('The installed version of OVS does not support '
                  'QoS rules for direct ports. '
                  'Please update to version %s or newer.',
                  checks.get_ovs_version_for_qos_direct_port_support())
    return result


def check_dnsmasq_local_service_supported():
    result = checks.dnsmasq_local_service_supported()
    if not result:
        LOG.error('The installed version of dnsmasq is too old. '
                  'Please update to a version supporting the '
                  '--local-service option.')
    return result


def check_keepalived_ipv6_support():
    result = checks.keepalived_ipv6_supported()
    if not result:
        LOG.error('The installed version of keepalived does not support '
                  'IPv6. Please update to at least version 1.2.10 for '
                  'IPv6 support.')
    return result


def check_keepalived_garp_on_sighup_support():
    result = checks.keepalived_garp_on_sighup_supported()
    if not result:
        LOG.error('The installed version of keepalived may not support '
                  'sending gratious ARP on SIGHUP, which may delay '
                  'dataplane downtime during HA router failover. '
                  'Please use at least version 1.2.20 which support '
                  'sending garp on SIGHUP.')
    return result


def check_arp_responder():
    result = checks.arp_responder_supported()
    if not result:
        LOG.error('Check for Open vSwitch ARP responder support failed. '
                  'Please ensure that the version of openvswitch '
                  'being used has ARP flows support.')
    return result


def check_arp_header_match():
    result = checks.arp_header_match_supported()
    if not result:
        LOG.error('Check for Open vSwitch support of ARP header matching '
                  'failed. ARP spoofing suppression will not work. A '
                  'newer version of OVS is required.')
    return result


def check_icmpv6_header_match():
    result = checks.icmpv6_header_match_supported()
    if not result:
        LOG.error('Check for Open vSwitch support of ICMPv6 header '
                  'matching failed. ICMPv6 Neighbor Advt spoofing (part '
                  'of arp spoofing) suppression will not work. A newer '
                  'version of OVS is required.')
    return result


def check_ovsdb_native():
    result = checks.ovsdb_native_supported()
    if not result:
        LOG.error('Check for native OVSDB support failed.')
    return result


def check_ovs_conntrack():
    result = checks.ovs_conntrack_supported()
    if not result:
        LOG.error('Check for Open vSwitch support of conntrack support '
                  'failed. OVS/CT firewall will not work. A newer '
                  'version of OVS (2.5+) and linux kernel (4.3+) are '
                  'required. See '
                  'https://github.com/openvswitch/ovs/blob/master/FAQ.md '
                  'for more information.')
    return result


def check_gre_conntrack():
    result = checks.gre_conntrack_supported()
    if not result:
        LOG.warning('Kernel module %s is not loaded. GRE tunnels from '
                    'VM to VM will not work with OVS firewall driver.',
                    checks.CONNTRACK_GRE_MODULE)
    return result


def check_ebtables():
    result = checks.ebtables_supported()
    if not result:
        LOG.error('Cannot run ebtables. Please ensure that it '
                  'is installed.')
    return result


def check_ipset():
    result = checks.ipset_supported()
    if not result:
        LOG.error('Cannot run ipset. Please ensure that it '
                  'is installed.')
    return result


def check_ip6tables():
    result = checks.ip6tables_supported()
    if not result:
        LOG.error('Cannot run ip6tables. Please ensure that it '
                  'is installed.')
    return result


def check_conntrack():
    result = checks.conntrack_supported()
    if not result:
        LOG.error('Cannot run conntrack. Please ensure that it '
                  'is installed.')
    return result


def check_dhcp_release6():
    result = checks.dhcp_release6_supported()
    if not result:
        LOG.error('No dhcp_release6 tool detected. The installed version '
                  'of dnsmasq does not support releasing IPv6 leases. '
                  'Please update to at least version %s if you need this '
                  'feature. If you do not use IPv6 stateful subnets you '
                  'can continue to use this version of dnsmasq, as '
                  'other IPv6 address assignment mechanisms besides '
                  'stateful DHCPv6 should continue to work without '
                  'the dhcp_release6 utility. '
                  'Current version of dnsmasq is ok if other checks '
                  'pass.',
                  checks.get_dnsmasq_version_with_dhcp_release6())
    return result


def check_bridge_firewalling_enabled():
    result = checks.bridge_firewalling_enabled()
    if not result:
        LOG.error('Bridge firewalling is not enabled. It may be the case '
                  'that bridge and/or br_netfilter kernel modules are not '
                  'loaded. Alternatively, corresponding sysctl settings '
                  'may be overridden to disable it by default.')
    return result


def check_ip_nonlocal_bind():
    result = checks.ip_nonlocal_bind()
    if not result:
        LOG.error('This kernel does not isolate ip_nonlocal_bind kernel '
                  'option in namespaces. Please update to kernel '
                  'version > 3.19.')
    return result


def check_min_tx_rate_support():
    result = checks.min_tx_rate_support()
    if not result:
        LOG.warning('There are SR-IOV network interfaces that do not support '
                    'setting the minimum TX rate (dataplane enforced minimum '
                    'guaranteed bandwidth) "ip-link vf min_tx_rate".')
    return result


def check_ovn_nb_db_schema_port_group_support():
    result = checks.ovn_nb_db_schema_port_group_supported()
    if not result:
        LOG.warning('OVN NB DB schema does not support Port_Group. This '
                    'support was added in DB schema version 5.11.')
    return result


def check_ovn_nb_db_schema_stateless_nat():
    result = checks.ovn_nb_db_schema_stateless_nat_supported()
    if not result:
        LOG.warning('OVN NB DB schema does not support stateless NAT. This '
                    'support was added in DB schema version 5.17.')
    return result


def check_ovn_sb_db_schema_virtual_port():
    result = checks.ovn_sb_db_schema_virtual_port_supported()
    if not result:
        LOG.warning('OVN SB DB schema does not support virtual ports. This '
                    'support was added in DB schema version 2.5.')
    return result


def check_ovn_localnet_learn_fdb_support():
    result = checks.ovn_localnet_learn_fdb_support()
    if not result:
        LOG.warning('OVN does not support localnet_learn_fdb option. '
                    'This support was added in OVN 22.09.')


def check_ovn_sb_db_schema_chassis_private():
    result = checks.ovn_sb_db_schema_chassis_private_supported()
    if not result:
        LOG.warning('OVN SB DB schema does not support chassis private. This '
                    'support was added in DB schema version 2.9.0.')
    return result


# Define CLI opts to test specific features, with a callback for the test
OPTS = [
    BoolOptCallback('ovs_vxlan', check_ovs_vxlan, default=False,
                    help=_('Check for OVS vxlan support')),
    BoolOptCallback('ovs_geneve', check_ovs_geneve, default=False,
                    help=_('Check for OVS Geneve support')),
    BoolOptCallback('iproute2_vxlan', check_iproute2_vxlan, default=False,
                    help=_('Check for iproute2 vxlan support')),
    BoolOptCallback('ovs_patch', check_ovs_patch, default=True,
                    help=_('Check for patch port support')),
    BoolOptCallback('arp_responder', check_arp_responder,
                    help=_('Check for ARP responder support')),
    BoolOptCallback('arp_header_match', check_arp_header_match,
                    help=_('Check for ARP header match support')),
    BoolOptCallback('icmpv6_header_match', check_icmpv6_header_match,
                    help=_('Check for ICMPv6 header match support')),
    BoolOptCallback('read_netns', check_read_netns,
                    help=_('Check netns permission settings')),
    BoolOptCallback('dnsmasq_local_service_supported',
                    check_dnsmasq_local_service_supported,
                    help=_('Check for local-service support in dnsmasq')),
    BoolOptCallback('ovs_qos_direct_port_supported',
                    check_ovs_qos_direct_ports_supported,
                    help=_('Check if the ovs supports QoS for direct ports')),
    BoolOptCallback('dnsmasq_version', check_dnsmasq_version,
                    help=_('Check minimal dnsmasq version'),
                    deprecated_for_removal=True,
                    deprecated_since='Pike'),
    BoolOptCallback('ovsdb_native', check_ovsdb_native,
                    help=_('Check ovsdb native interface support')),
    BoolOptCallback('ovs_conntrack', check_ovs_conntrack,
                    help=_('Check ovs conntrack support')),
    BoolOptCallback('gre_conntrack', check_gre_conntrack,
                    help=_('Check if conntrack for gre tunnels traffic is '
                           'supported')),
    BoolOptCallback('ebtables_installed', check_ebtables,
                    help=_('Check ebtables installation')),
    BoolOptCallback('keepalived_ipv6_support', check_keepalived_ipv6_support,
                    help=_('Check keepalived IPv6 support')),
    BoolOptCallback('keepalived_garp_on_sighup_support',
                    check_keepalived_garp_on_sighup_support,
                    help=_('Check keepalived support sending garp on '
                           'SIGHUP.')),
    BoolOptCallback('ipset_installed', check_ipset,
                    help=_('Check ipset installation')),
    BoolOptCallback('ip6tables_installed', check_ip6tables,
                    help=_('Check ip6tables installation')),
    BoolOptCallback('conntrack_installed', check_conntrack,
                    help=_('Check conntrack installation')),
    BoolOptCallback('dhcp_release6', check_dhcp_release6,
                    help=_('Check dhcp_release6 installation')),
    BoolOptCallback('bridge_firewalling', check_bridge_firewalling_enabled,
                    help=_('Check bridge firewalling'),
                    default=False),
    BoolOptCallback('ip_nonlocal_bind', check_ip_nonlocal_bind,
                    help=_('Check ip_nonlocal_bind kernel option works with '
                           'network namespaces.'),
                    default=False),
    BoolOptCallback('check_min_tx_rate_support', check_min_tx_rate_support,
                    help=_('Check if the configured SR-IOV NICs support '
                           'the "ip-link vf min_tx_rate" parameter.'),
                    default=False),
    BoolOptCallback('ovn_nb_db_schema_port_group_support',
                    check_ovn_nb_db_schema_port_group_support,
                    help=_('Check OVN NB DB schema support Port_Group'),
                    default=False),
    BoolOptCallback('ovn_nb_db_schema_stateless_nat_support',
                    check_ovn_nb_db_schema_stateless_nat,
                    help=_('Check OVN NB DB schema support stateless NAT'),
                    default=False),
    BoolOptCallback('ovn_sb_db_schema_virtual_port_support',
                    check_ovn_sb_db_schema_virtual_port,
                    help=_('Check OVN SB DB schema support virtual ports'),
                    default=False),
    BoolOptCallback('ovn_localnet_learn_fdb_support',
                    check_ovn_localnet_learn_fdb_support,
                    help=_('Check OVN supports localnet_learn_fdb option'),
                    default=False),
    BoolOptCallback('ovn_sb_db_schema_chassis_private_support',
                    check_ovn_sb_db_schema_chassis_private,
                    help=_('Check OVN SB DB schema supports Chassis_Private'),
                    default=False),
]


def enable_tests_from_config():
    """If a test can depend on configuration, use this function to set the
    appropriate CLI option to enable that test. It will then be possible to
    run all necessary tests, just by passing in the appropriate configs.
    """

    cfg.CONF.set_default('arp_header_match', True)
    cfg.CONF.set_default('icmpv6_header_match', True)
    if constants.TYPE_VXLAN in cfg.CONF.AGENT.tunnel_types:
        cfg.CONF.set_default('ovs_vxlan', True)
    if constants.TYPE_GENEVE in cfg.CONF.AGENT.tunnel_types:
        cfg.CONF.set_default('ovs_geneve', True)
    if (constants.TYPE_VXLAN in cfg.CONF.ml2.type_drivers or
            cfg.CONF.VXLAN.enable_vxlan):
        cfg.CONF.set_default('iproute2_vxlan', True)
    if cfg.CONF.AGENT.arp_responder:
        cfg.CONF.set_default('arp_responder', True)
    if not cfg.CONF.AGENT.use_helper_for_ns_read:
        cfg.CONF.set_default('read_netns', True)
    if cfg.CONF.dhcp_driver == 'neutron.agent.linux.dhcp.Dnsmasq':
        cfg.CONF.set_default('dnsmasq_local_service_supported', True)
        cfg.CONF.set_default('dnsmasq_version', True)
    if cfg.CONF.l3_ha:
        cfg.CONF.set_default('keepalived_ipv6_support', True)
        cfg.CONF.set_default('ip_nonlocal_bind', True)
    if cfg.CONF.SECURITYGROUP.enable_ipset:
        cfg.CONF.set_default('ipset_installed', True)
    if cfg.CONF.SECURITYGROUP.enable_security_group:
        cfg.CONF.set_default('ip6tables_installed', True)
    if cfg.CONF.SECURITYGROUP.firewall_driver in (
            'iptables',
            'iptables_hybrid',
            ('neutron.agent.linux.iptables_firewall.'
             'IptablesFirewallDriver'),
            ('neutron.agent.linux.iptables_firewall.'
             'OVSHybridIptablesFirewallDriver'),
    ):
        cfg.CONF.set_default('bridge_firewalling', True)
    if cfg.CONF.SRIOV_NIC.physical_device_mappings:
        cfg.CONF.set_default('check_min_tx_rate_support', True)
    if 'ovn' in cfg.CONF.ml2.mechanism_drivers:
        cfg.CONF.set_default('ovn_nb_db_schema_port_group_support', True)
        cfg.CONF.set_default('ovn_nb_db_schema_stateless_nat_support', True)
        cfg.CONF.set_default('ovn_sb_db_schema_virtual_port_support', True)


def all_tests_passed():
    return all(opt.callback() for opt in OPTS if cfg.CONF.get(opt.name))


def main():
    setup_conf()
    cfg.CONF.register_cli_opts(OPTS)
    cfg.CONF.set_override('use_stderr', True)
    config.setup_logging()
    config.init(sys.argv[1:], default_config_files=[])
    agent_config.setup_privsep()

    if cfg.CONF.config_file:
        enable_tests_from_config()

    return 0 if all_tests_passed() else 1



# Copyright (c) 2015 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re
import signal

from neutron_lib import constants
from neutron_lib.plugins.ml2 import ovs_constants
from oslo_log import log as logging

from neutron.agent.common import async_process
from neutron.common import utils as common_utils
from neutron.plugins.ml2.common import constants as comm_consts

LOG = logging.getLogger(__name__)

PACKET_RATE_LIMIT = ovs_constants.PACKET_RATE_LIMIT
BANDWIDTH_RATE_LIMIT = ovs_constants.BANDWIDTH_RATE_LIMIT


class TcpdumpException(Exception):
    pass


def extract_mod_nw_tos_action(flows):
    tos_mark = None
    if flows:
        flow_list = flows.splitlines()
        for flow in flow_list:
            if 'mod_nw_tos' in flow:
                actions = flow.partition('actions=')[2]
                after_mod = actions.partition('mod_nw_tos:')[2]
                tos_mark = int(after_mod.partition(',')[0])
    return tos_mark


def wait_until_bandwidth_limit_rule_applied(check_function, port_vif, rule):
    def _bandwidth_limit_rule_applied():
        bw_rule = check_function(port_vif)
        expected = None, None
        if rule:
            expected = rule.max_kbps, rule.max_burst_kbps
        return bw_rule == expected

    common_utils.wait_until_true(_bandwidth_limit_rule_applied)


def wait_until_egress_bandwidth_limit_rule_applied(bridge, port_vif, rule):
    wait_until_bandwidth_limit_rule_applied(
        bridge.get_egress_bw_limit_for_port, port_vif, rule)


def wait_until_ingress_bandwidth_limit_rule_applied(bridge, port_vif, rule):
    wait_until_bandwidth_limit_rule_applied(
        bridge.get_ingress_bw_limit_for_port, port_vif, rule)


def wait_until_dscp_marking_rule_applied_ovs(bridge, port_vif, rule):
    def _dscp_marking_rule_applied():
        port_num = bridge.get_port_ofport(port_vif)

        flows = bridge.dump_flows_for(table='0', in_port=str(port_num))
        dscp_mark = extract_mod_nw_tos_action(flows)

        expected = None
        if rule:
            expected = rule << 2
        return dscp_mark == expected

    common_utils.wait_until_true(_dscp_marking_rule_applied)


def wait_until_pkt_meter_rule_applied_ovs(bridge, port_vif, port_id,
                                          direction, mac=None,
                                          type_=comm_consts.METER_FLAG_PPS):
    def _pkt_rate_limit_rule_applied():
        port_num = bridge.get_port_ofport(port_vif)
        port_vlan = bridge.get_port_tag_by_name(port_vif)
        key = f"{type_}_{port_id}_{direction}"
        meter_id = bridge.get_value_from_other_config(
            port_vif, key, value_type=int)

        table = (str(PACKET_RATE_LIMIT) if
                 type_ == comm_consts.METER_FLAG_PPS else
                 str(BANDWIDTH_RATE_LIMIT))

        if direction == constants.EGRESS_DIRECTION:
            flows = bridge.dump_flows_for(table=table, in_port=str(port_num),
                                          dl_src=str(mac))
        else:
            flows = bridge.dump_flows_for(table=table, dl_vlan=str(port_vlan),
                                          dl_dst=str(mac))
        if mac:
            return bool(flows) and meter_id
        return not bool(flows) and not meter_id

    common_utils.wait_until_true(_pkt_rate_limit_rule_applied)


def wait_for_dscp_marked_packet(sender_vm, receiver_vm, dscp_mark):
    cmd = [
        "tcpdump", "-i", receiver_vm.port.name, "-nlt",
        "src", sender_vm.ip, 'and', 'dst', receiver_vm.ip]
    if dscp_mark:
        cmd += ["and", "(ip[1] & 0xfc == %s)" % (dscp_mark << 2)]
    tcpdump_async = async_process.AsyncProcess(cmd, run_as_root=True,
                                               namespace=receiver_vm.namespace)
    tcpdump_async.start(block=True)

    sender_vm.block_until_ping(receiver_vm.ip)

    try:
        tcpdump_async.stop(kill_signal=signal.SIGINT)
    except async_process.AsyncProcessException:
        # If it was already stopped than we don't care about it
        pass

    tcpdump_stderr_lines = []
    pattern = r"(?P<packets_count>^\d+) packets received by filter"
    for line in tcpdump_async.iter_stderr():
        m = re.match(pattern, line)
        if m and int(m.group("packets_count")) != 0:
            return
        tcpdump_stderr_lines.append(line)

    tcpdump_stdout_lines = list(tcpdump_async.iter_stdout())
    LOG.debug("Captured output lines from tcpdump. Stdout: %s; Stderr: %s",
              tcpdump_stdout_lines, tcpdump_stderr_lines)

    raise TcpdumpException(
        "No packets marked with DSCP = %(dscp_mark)s received from %(src)s "
        "to %(dst)s" % {'dscp_mark': dscp_mark,
                        'src': sender_vm.ip,
                        'dst': receiver_vm.ip})


def extract_vlan_id(flows):
    if flows:
        flow_list = flows.splitlines()
        for flow in flow_list:
            if 'mod_vlan_vid' in flow:
                actions = flow.partition('actions=')[2]
                after_mod = actions.partition('mod_vlan_vid:')[2]
                return int(after_mod.partition(',')[0])


def extract_ofports(flows):
    if flows:
        flow_list = flows.splitlines()
        for flow in flow_list:
            if 'dl_vlan' in flow:
                actions = flow.partition('actions=')[2]
                outputs = set()
                for action in actions.split(','):
                    if 'output' in action:
                        outputs.add(int(action.partition(':')[2]))
                return outputs


def extract_br_tun_cookie(bridge):
    flows = bridge.dump_flows_for_table(ovs_constants.FLOOD_TO_TUN)
    if flows:
        flow_list = flows.splitlines()
        for flow in flow_list:
            if 'cookie' in flow:
                cookie = flow.split(',')[0].partition('cookie=')[2]
                return cookie


def wait_for_mod_vlan_id_applied(bridge, expected_vlan_id):
    def _vlan_id_rule_applied():
        flows = bridge.dump_flows_for_table(0)
        vlan_id = extract_vlan_id(flows)
        return vlan_id == expected_vlan_id

    common_utils.wait_until_true(_vlan_id_rule_applied)


def wait_for_br_tun_flow_added(bridge, expected_output):
    def _lookup():
        flows = bridge.dump_flows_for_table(ovs_constants.FLOOD_TO_TUN)
        output = extract_ofports(flows)
        return output == expected_output

    common_utils.wait_until_true(_lookup)


def wait_for_br_tun_cleanup_flows(bridge, cookie):
    def _lookup():
        flows = bridge.dump_flows_for_table(ovs_constants.FLOOD_TO_TUN)
        if flows:
            flow_list = flows.splitlines()
            for flow in flow_list:
                # Return false until we get rid of
                # dl_vlan flows
                if cookie in flow:
                    return False
        return True

    common_utils.wait_until_true(_lookup)



# Copyright 2024 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import rpc as n_rpc
from oslo_messaging import Target


class BasePluginApi:
    """Base agent side of the rpc API"""
    def __init__(self, topic, namespace, version):
        target = Target(
            topic=topic,
            namespace=namespace,
            version=version)
        self.client = n_rpc.get_client(target)

    def get_ports(self, context, port_filters):
        # NOTE(mtomaska): The MetadataRpcCallback (server side) API version 1.0
        # exposes get_ports, under the PLUGIN topic and None namespace.
        cctxt = self.client.prepare(version='1.0')
        return cctxt.call(context, 'get_ports', filters=port_filters)



# Copyright 2025 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.callbacks import registry
from neutron_lib.services import base as service_base
from oslo_config import cfg
from oslo_log import log

from neutron.conf.services import bgp as bgp_config
from neutron.services.bgp import worker

LOG = log.getLogger(__name__)


@registry.has_registry_receivers
class BGPServicePlugin(service_base.ServicePluginBase):

    supported_extension_aliases = []

    def __init__(self):
        LOG.info("Starting BGP Service Plugin")
        super().__init__()
        bgp_config.register_opts(cfg.CONF)

    def get_workers(self):
        return [worker.BGPWorker()]

    def get_plugin_description(self):
        return "BGP service plugin for OVN"

    @classmethod
    def get_plugin_type(cls):
        return "bgp-service"



# Copyright (c) 2018 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils

from neutron.agent.common import ovs_lib
from neutron.common import utils as common_utils
from neutron.tests.functional import base


class BridgeMonitorTestCase(base.BaseSudoTestCase):

    def _delete_bridges(self, bridges):
        for bridge in bridges:
            self.ovs.delete_bridge(bridge)

    def test_create_bridges(self):
        bridges_to_create = [
            'br_' + uuidutils.generate_uuid()[:12],
            'br_' + uuidutils.generate_uuid()[:12],
            'br_' + uuidutils.generate_uuid()[:12],
            'br_' + uuidutils.generate_uuid()[:12],
            'br_' + uuidutils.generate_uuid()[:12],
        ]
        bridges_to_monitor = bridges_to_create[:3]
        self.ovs = ovs_lib.BaseOVS()
        self._delete_bridges(bridges_to_create)

        self.ovs.ovsdb.idl_monitor.start_bridge_monitor(bridges_to_monitor)
        self.addCleanup(self._delete_bridges, bridges_to_create)
        for bridge in bridges_to_create:
            self.ovs.add_bridge(bridge)

        _idl_mon = self.ovs.ovsdb.idl_monitor
        try:
            common_utils.wait_until_true(
                lambda: set(bridges_to_monitor) ==
                        set(_idl_mon._bridges_added_list),
                timeout=10)
        except common_utils.WaitTimeout:
            self.fail(f'Bridges to monitor: {set(bridges_to_monitor)}, '
                      f'bridges added: {set(_idl_mon._bridges_added_list)}')
        self.assertEqual(bridges_to_monitor, _idl_mon.bridges_added)
        self.assertEqual([], _idl_mon.bridges_added)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import collections
import copy
import itertools
import typing

from neutron_lib.api import attributes
from neutron_lib.api.definitions import port
from neutron_lib.api import extensions as api_extensions
from neutron_lib.api import faults
from neutron_lib.api import validators
from neutron_lib import constants
from neutron_lib.db import standard_attr
from neutron_lib import exceptions
from neutron_lib.plugins import directory
from neutron_lib import rpc as n_rpc
from neutron_lib.services import base as service_base

from neutron._i18n import _
from neutron.api import extensions
from neutron.api.v2 import resource as api_resource
from neutron.objects import network as network_obj
from neutron.objects import network_segment_range as network_segment_range_obj
from neutron.objects import ports as ports_obj
from neutron.objects.qos import policy as policy_obj
from neutron.objects import router as router_obj
from neutron.objects import securitygroup as securitygroup_obj
from neutron.objects import subnet as subnet_obj
from neutron.objects import subnetpool as subnetpool_obj
from neutron.objects import trunk as trunk_obj
from neutron import policy


TAG = 'tag'
TAGS = TAG + 's'
TAGS_ANY = TAGS + '-any'
NOT_TAGS = 'not-' + TAGS
NOT_TAGS_ANY = NOT_TAGS + '-any'
MAX_TAG_LEN = 255
MAX_TAGS_COUNT = 50
TAG_PLUGIN_TYPE = 'TAG'

TAG_SUPPORTED_RESOURCES = standard_attr.get_tag_resource_parent_map()
TAG_ATTRIBUTE_MAP = {
    TAGS: {'allow_post': False, 'allow_put': False,
           'is_visible': True, 'is_filter': True},
    TAGS_ANY: {'allow_post': False, 'allow_put': False,
               'is_visible': False, 'is_filter': True},
    NOT_TAGS: {'allow_post': False, 'allow_put': False,
               'is_visible': False, 'is_filter': True},
    NOT_TAGS_ANY: {'allow_post': False, 'allow_put': False,
                   'is_visible': False, 'is_filter': True},
}

TAG_ATTRIBUTE_MAP_PORTS: dict[str, typing.Any]
TAG_ATTRIBUTE_MAP_PORTS = copy.deepcopy(TAG_ATTRIBUTE_MAP)
TAG_ATTRIBUTE_MAP_PORTS[TAGS] = {
        'allow_post': True, 'allow_put': False,
        'validate': {'type:list_of_unique_strings': MAX_TAG_LEN},
        'default': [], 'is_visible': True, 'is_filter': True
}
OVO_CLS = {
    'floatingips': router_obj.FloatingIP,
    'network_segment_ranges': network_segment_range_obj.NetworkSegmentRange,
    'networks': network_obj.Network,
    'policies': policy_obj.QosPolicy,
    'ports': ports_obj.Port,
    'routers': router_obj.Router,
    'security_groups': securitygroup_obj.SecurityGroup,
    'subnets': subnet_obj.Subnet,
    'subnetpools': subnetpool_obj.SubnetPool,
    'trunks': trunk_obj.Trunk,
}
ResourceInfo = collections.namedtuple(
    'ResourceInfo', ['project_id',
                     'obj_type',
                     'obj',
                     ])
EMPTY_RESOURCE_INFO = ResourceInfo(None, None, None)


class TagResourceNotFound(exceptions.NotFound):
    message = _("Resource %(resource)s %(resource_id)s could not be found.")


class TagNotFound(exceptions.NotFound):
    message = _("Tag %(tag)s could not be found.")


def validate_tag(tag):
    msg = validators.validate_string(tag, MAX_TAG_LEN)
    if msg:
        raise exceptions.InvalidInput(error_message=msg)


def validate_tags_limit(resource, tags):
    tags = set(tags)
    if len(tags) > MAX_TAGS_COUNT:
        msg = (_("The number of tags exceed the per-resource limit of %d")
               % MAX_TAGS_COUNT)
        raise exceptions.BadRequest(resource=resource, msg=msg)


def validate_tags(body):
    if not isinstance(body, dict) or 'tags' not in body:
        raise exceptions.InvalidInput(error_message=_("Invalid tags body"))
    msg = validators.validate_list_of_unique_strings(body['tags'], MAX_TAG_LEN)
    if msg:
        raise exceptions.InvalidInput(error_message=msg)


def notify_tag_action(context, action, obj, obj_id, tags=None):
    notifier = n_rpc.get_notifier('network')
    tag_event = 'tag.%s' % action
    # TODO(hichihara): Add 'updated_at' into payload
    payload = {'obj_resource': obj,
               'obj_resource_id': obj_id}
    if tags is not None:
        payload['tags'] = tags
    notifier.info(context, tag_event, payload)


class TaggingController:
    def __init__(self):
        self.plugin = directory.get_plugin(TAG_PLUGIN_TYPE)
        self.supported_resources = TAG_SUPPORTED_RESOURCES

    def _get_resource_info(self, context, kwargs, tags=None):
        """Return the information about the resource with the tag(s)

        :param kwargs: dictionary with the resource ID, along with other
                       information. It is formated as
                       {"resource_id": "id", ...}
        :param tags: list of the tags which will be set for the resource
        :return: ``ResourceInfo`` named tuple with the object's type,
                 object's information in the dict and the project ID
        """
        for key, obj_type in itertools.product(
                kwargs.keys(), self.supported_resources.keys()):
            if key != self.supported_resources[obj_type] + '_id':
                continue

            obj_id = kwargs[key]
            obj_class = OVO_CLS[obj_type]
            try:
                field_list = []
                for attr_name, attr_config in \
                        attributes.RESOURCES[obj_type].items():
                    if (attr_config.get('required_by_policy') or
                            attr_config.get('primary_key') or
                            'default' not in attr_config):
                        field_list.append(attr_name)
                obj_dict = {
                    constants.ATTRIBUTES_TO_UPDATE: [TAGS]
                }
                if tags is not None:
                    obj_dict[TAGS] = tags
                obj = obj_class.get_object(context.elevated(), id=obj_id,
                                           fields=field_list)
                if not obj:
                    return EMPTY_RESOURCE_INFO
                for f_name, f_value in obj.to_dict().items():
                    if f_name in field_list:
                        obj_dict[f_name] = f_value
                project_id = obj_dict.get('project_id')
                if not project_id:
                    project_id = obj_dict.get('tenant_id')
                    obj_dict['project_id'] = project_id
            except IndexError:
                return EMPTY_RESOURCE_INFO

            return ResourceInfo(project_id, obj_type, obj_dict)

        # This should never be returned.
        return EMPTY_RESOURCE_INFO

    def _get_policy_action(self, base_action, obj_type):
        return "{}_{}:{}".format(
            base_action,
            self.supported_resources[obj_type],
            TAGS)

    def index(self, request, **kwargs):
        # GET /v2.0/{obj_resource}/{obj_resource_id}/tags
        ctx = request.context
        rinfo = self._get_resource_info(ctx, kwargs)
        policy.enforce(ctx, f'get_{rinfo.obj_type}_{TAGS}',
                       rinfo.obj)
        return self.plugin.get_tags(ctx, rinfo.obj_type, rinfo.obj['id'])

    def show(self, request, id, **kwargs):
        # GET /v2.0/{obj_resource}/{obj_resource_id}/tags/{tag}
        # id == tag
        validate_tag(id)
        ctx = request.context
        rinfo = self._get_resource_info(ctx, kwargs)
        policy.enforce(ctx, f'get_{rinfo.obj_type}:{TAGS}',
                       rinfo.obj)
        return self.plugin.get_tag(ctx, rinfo.obj_type, rinfo.obj['id'], id)

    def create(self, request, body, **kwargs):
        # POST /v2.0/{obj_resource}/{obj_resource_id}/tags
        # body: {"tags": ["aaa", "bbb"]}
        validate_tags(body)
        ctx = request.context
        rinfo = self._get_resource_info(ctx, kwargs, tags=body[TAGS])
        policy.enforce(ctx, f'create_{rinfo.obj_type}:{TAGS}',
                       rinfo.obj)
        validate_tags_limit(rinfo.obj_type, body['tags'])
        notify_tag_action(ctx, 'create.start', rinfo.obj_type,
                          rinfo.obj['id'], body['tags'])
        result = self.plugin.create_tags(ctx, rinfo.obj_type,
                                         rinfo.obj['id'], body)
        notify_tag_action(ctx, 'create.end', rinfo.obj_type,
                          rinfo.obj['id'], body['tags'])
        return result

    def update(self, request, id, **kwargs):
        # PUT /v2.0/{obj_resource}/{obj_resource_id}/tags/{tag}
        # id == tag
        validate_tag(id)
        ctx = request.context
        rinfo = self._get_resource_info(ctx, kwargs, tags=[id])
        policy.enforce(ctx, f'update_{rinfo.obj_type}:{TAGS}',
                       rinfo.obj)
        current_tags = self.plugin.get_tags(
            ctx, rinfo.obj_type, rinfo.obj['id'])['tags']
        new_tags = current_tags + [id]
        validate_tags_limit(rinfo.obj_type, new_tags)
        notify_tag_action(ctx, 'create.start', rinfo.obj_type,
                          rinfo.obj['id'], [id])
        result = self.plugin.update_tag(ctx, rinfo.obj_type,
                                        rinfo.obj['id'], id)
        notify_tag_action(ctx, 'create.end', rinfo.obj_type,
                          rinfo.obj['id'], [id])
        return result

    def update_all(self, request, body, **kwargs):
        # PUT /v2.0/{obj_resource}/{obj_resource_id}/tags
        # body: {"tags": ["aaa", "bbb"]}
        validate_tags(body)
        ctx = request.context
        rinfo = self._get_resource_info(ctx, kwargs, tags=body[TAGS])
        policy.enforce(
            ctx,
            self._get_policy_action("update", rinfo.obj_type),
            rinfo.obj)
        validate_tags_limit(rinfo.obj_type, body['tags'])
        notify_tag_action(ctx, 'update.start', rinfo.obj_type,
                          rinfo.obj['id'], body['tags'])
        result = self.plugin.update_tags(ctx, rinfo.obj_type,
                                         rinfo.obj['id'], body)
        notify_tag_action(ctx, 'update.end', rinfo.obj_type,
                          rinfo.obj['id'], body['tags'])
        return result

    def delete(self, request, id, **kwargs):
        # DELETE /v2.0/{obj_resource}/{obj_resource_id}/tags/{tag}
        # id == tag
        validate_tag(id)
        ctx = request.context
        rinfo = self._get_resource_info(ctx, kwargs)
        policy.enforce(
            ctx,
            self._get_policy_action("delete", rinfo.obj_type),
            rinfo.obj)
        notify_tag_action(ctx, 'delete.start', rinfo.obj_type,
                          rinfo.obj['id'], [id])
        result = self.plugin.delete_tag(ctx, rinfo.obj_type,
                                        rinfo.obj['id'], id)
        notify_tag_action(ctx, 'delete.end', rinfo.obj_type,
                          rinfo.obj['id'], [id])
        return result

    def delete_all(self, request, **kwargs):
        # DELETE /v2.0/{obj_resource}/{obj_resource_id}/tags
        ctx = request.context
        rinfo = self._get_resource_info(ctx, kwargs)
        policy.enforce(
            ctx,
            self._get_policy_action("delete", rinfo.obj_type),
            rinfo.obj)
        notify_tag_action(ctx, 'delete_all.start', rinfo.obj_type,
                          rinfo.obj['id'])
        result = self.plugin.delete_tags(ctx, rinfo.obj_type,
                                         rinfo.obj['id'])
        notify_tag_action(ctx, 'delete_all.end', rinfo.obj_type,
                          rinfo.obj['id'])
        return result


class Tagging(api_extensions.ExtensionDescriptor):
    """Extension class supporting tags."""

    @classmethod
    def get_name(cls):
        return ("Tag support for resources with standard attribute: %s"
                % ', '.join(TAG_SUPPORTED_RESOURCES.values()))

    @classmethod
    def get_alias(cls):
        return "standard-attr-tag"

    @classmethod
    def get_description(cls):
        return "Enables to set tag on resources with standard attribute."

    @classmethod
    def get_updated(cls):
        return "2017-01-01T00:00:00-00:00"

    def get_required_extensions(self):
        return []

    @classmethod
    def get_resources(cls):
        """Returns Ext Resources."""
        exts = []
        action_status = {'index': 200, 'show': 204, 'update': 201,
                         'update_all': 200, 'delete': 204, 'delete_all': 204}
        controller = api_resource.Resource(TaggingController(),
                                           faults.FAULT_MAP,
                                           action_status=action_status)
        collection_methods = {"delete_all": "DELETE",
                              "update_all": "PUT"}
        exts = []
        for collection_name, member_name in TAG_SUPPORTED_RESOURCES.items():
            if 'security_group' in collection_name:
                collection_name = collection_name.replace('_', '-')
            obj = {'member_name': member_name,
                   'collection_name': collection_name}
            exts.append(extensions.ResourceExtension(
                TAGS, controller, obj,
                collection_methods=collection_methods))
        return exts

    def get_extended_resources(self, version):
        if version != "2.0":
            return {}
        EXTENDED_ATTRIBUTES_2_0 = {}
        for collection_name in TAG_SUPPORTED_RESOURCES:
            if collection_name == port.COLLECTION_NAME:
                EXTENDED_ATTRIBUTES_2_0[collection_name] = (
                    TAG_ATTRIBUTE_MAP_PORTS)
            else:
                EXTENDED_ATTRIBUTES_2_0[collection_name] = TAG_ATTRIBUTE_MAP
        return EXTENDED_ATTRIBUTES_2_0


class TagPluginBase(service_base.ServicePluginBase, metaclass=abc.ABCMeta):
    """REST API to operate the Tag."""

    def get_plugin_description(self):
        return "Tag support"

    @classmethod
    def get_plugin_type(cls):
        return TAG_PLUGIN_TYPE

    @abc.abstractmethod
    def get_tags(self, context, resource, resource_id):
        pass

    @abc.abstractmethod
    def get_tag(self, context, resource, resource_id, tag):
        pass

    @abc.abstractmethod
    def update_tags(self, context, resource, resource_id, body):
        pass

    @abc.abstractmethod
    def update_tag(self, context, resource, resource_id, tag):
        pass

    @abc.abstractmethod
    def delete_tags(self, context, resource, resource_id):
        pass

    @abc.abstractmethod
    def delete_tag(self, context, resource, resource_id, tag):
        pass



# Copyright 2025 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
from neutron_lib import constants as n_const
from oslo_log import log
from ovsdbapp.backend.ovs_idl import command as ovs_cmd
from ovsdbapp.backend.ovs_idl import idlutils
from ovsdbapp import constants as ovsdbapp_const
from ovsdbapp.schema.ovn_northbound import commands as nb_cmd

from neutron.agent.linux import ip_lib
from neutron.common.ovn import constants as ovn_const
from neutron.conf.services import bgp as bgp_config
from neutron.services.bgp import constants
from neutron.services.bgp import exceptions
from neutron.services.bgp import helpers

LOG = log.getLogger(__name__)


def _run_idl_command(cmd, txn):
    """A wrapper around a command to run it and return the row result.

    This avoids using the self.result attribute but returns a custom row_result
    instead. Important is that in case of a new row creation, the row_result
    does not contain the same UUID that is stored in the DB after the commit.
    """
    cmd.run_idl(txn)
    try:
        return cmd.row_result
    except AttributeError:
        # If the command does not implement the row_result attribute,
        # return the result
        return cmd.result


def _get_lrp_peer_ip(nb_idl, lrp):
    peer_lrp = nb_idl.lookup('Logical_Router_Port', lrp.peer[0])
    try:
        return str(netaddr.IPNetwork(peer_lrp.networks[0]).ip)
    except IndexError:
        return helpers.ipv6_link_local_from_mac(peer_lrp.mac)


def _get_main_router(nb_idl):
    return nb_idl.lookup('Logical_Router', bgp_config.get_main_router_name())


def _get_all_provider_switches(nb_idl):
    return [
        s for s in nb_idl.tables['Logical_Switch'].rows.values()
        if hasattr(s, 'external_ids') and s.external_ids.get(
            ovn_const.OVN_NETTYPE_EXT_ID_KEY) in n_const.TYPE_PHYSICAL]


def _get_gw_ips_for_switch(nb_idl, switch):
    n_net_id = helpers.get_neutron_id_from_ovn_name(switch)
    gw_ips = [
        gw_ip for dhcp_opt in nb_idl.tables['DHCP_Options'].rows.values()
        if dhcp_opt.external_ids.get(
            ovn_const.OVN_NETWORK_ID_EXT_ID_KEY) == n_net_id and
        (gw_ip := helpers.get_gw_ip_from_dhcp_options(dhcp_opt)) is not None]
    LOG.debug("For logical switch %s, found gateway IPs: %s",
              switch.name, gw_ips)
    return gw_ips


class _LsAddCommand(nb_cmd.LsAddCommand):
    def run_idl(self, txn):
        try:
            self.row_result = self.result = self.api.lookup(
                'Logical_Switch', self.switch)
        except idlutils.RowNotFound:
            super().run_idl(txn)
            self.row_result = self.api.lookup('Logical_Switch', self.switch)
            return

        self.set_columns(self.row_result, **self.columns)


class _LrAddCommand(nb_cmd.LrAddCommand):
    """An idempotent command to add a logical router.

    We need to subclass the LrAddCommand because it does not check if the
    columns in the existing row are the same as the columns we are trying to
    set.
    """
    def run_idl(self, txn):
        try:
            self.row_result = self.result = self.api.lookup(
                'Logical_Router', self.router)
        except idlutils.RowNotFound:
            super().run_idl(txn)
            self.row_result = self.api.lookup('Logical_Router', self.router)
            return

        self.set_columns(self.row_result, **self.columns)


class _LspAddCommand(nb_cmd.LspAddCommand):
    def run_idl(self, txn):
        try:
            self.row_result = self.result = self.api.lookup(
                'Logical_Switch_Port', self.port)
        except idlutils.RowNotFound:
            super().run_idl(txn)
            self.row_result = self.api.lookup('Logical_Switch_Port', self.port)
            return

        self.set_columns(self.row_result, **self.columns)


class _LrpAddCommand(nb_cmd.LrpAddCommand):
    """An idempotent command to add a logical router port.

    We need to subclass the LrpAddCommand because it does not check if the
    columns in the existing row are the same as the columns we are trying to
    set.
    """
    def __init__(
            self, api, router_name, lrp_name, networks=None, **kwargs):
        networks = networks or []
        mac = helpers.get_mac_address_from_lrp_name(lrp_name)
        super().__init__(api, router_name, lrp_name, mac, networks, **kwargs)

    def run_idl(self, txn):
        try:
            self.row_result = self.result = self.api.lookup(
                'Logical_Router_Port', self.port)
        except idlutils.RowNotFound:
            super().run_idl(txn)
            self.row_result = self.api.lookup('Logical_Router_Port', self.port)
            return

        # TODO(jlibosva): Make sure the mac is unique for this router
        self.row_result.mac = self.mac
        self.row_result.networks = self.networks
        self.row_result.peer = self.peer
        self.set_columns(self.row_result, **self.columns)


class _HAChassisGroupAddCommand(nb_cmd.HAChassisGroupAddCommand):
    """An idempotent command to add a HA chassis group.

    We need to subclass the HAChassisGroupAddCommand because it does not check
    if the columns in the existing row are the same as the columns we are
    trying to set.
    """
    def run_idl(self, txn):
        try:
            self.row_result = self.result = self.api.lookup(
                'HA_Chassis_Group', self.name)
        except idlutils.RowNotFound:
            super().run_idl(txn)
            self.row_result = self.api.lookup(
                'HA_Chassis_Group', self.name)
            return

        self.set_columns(self.row_result, **self.columns)


class _LrPolicyAddCommand(nb_cmd.LrPolicyAddCommand):
    """An idempotent command to add a logical router policy.
    """
    def __init__(self, api, router, priority, match, action, output_port,
                 *args, **kwargs):

        if 'nexthops' not in kwargs:
            kwargs['nexthops'] = [ip_lib.get_ipv6_lladdr(output_port.mac)]

        super().__init__(api, router, priority, match, action,
                         may_exist=True, output_port=output_port,
                         *args, **kwargs)


class CreateSwitchWithLocalnetCommand(_LsAddCommand):
    def __init__(self, api, name, network_name):
        super().__init__(api, name, may_exist=True)
        self.network_name = network_name

    def run_idl(self, txn):
        super().run_idl(txn)

        CreateLspLocalnetCommand(
            self.api, self.switch, self.network_name,
        ).run_idl(txn)


class CreateLspLocalnetCommand(_LspAddCommand):
    def __init__(self, api, switch_name, network_name):
        localnet_lsp_name = helpers.get_lsp_localnet_name(switch_name)
        super().__init__(
            api, switch_name, localnet_lsp_name, may_exist=True)
        self.network_name = network_name

    def run_idl(self, txn):
        self.columns = {
            'type': ovn_const.LSP_TYPE_LOCALNET,
            'options': {'network_name': self.network_name},
            'addresses': [ovn_const.UNKNOWN_ADDR],
        }
        super().run_idl(txn)


class ReconcileNeutronSwitchCommand(ovs_cmd.BaseCommand):
    def __init__(self, api, n_switch):
        super().__init__(api)
        self.n_switch = n_switch
        self.router_name = bgp_config.get_main_router_name()
        self.network_name = self._get_network_name()
        self.interconnect_switch_name = (
            helpers.get_provider_interconnect_switch_name(self.n_switch.name))

    def _get_network_name(self):
        for port in self.n_switch.ports:
            if port.type == ovn_const.LSP_TYPE_LOCALNET:
                return port.options['network_name']
        raise ValueError(
            f"No localnet port found for switch {self.n_switch.name}")

    def run_idl(self, txn):
        # The fake connection to distribute the routes to the BGP router
        ConnectRouterToSwitchCommand(
            self.api,
            self.router_name,
            self.n_switch.name,
        ).run_idl(txn)

        CreateSwitchWithLocalnetCommand(
            self.api,
            self.interconnect_switch_name,
            self.network_name,
        ).run_idl(txn)

        ConnectRouterToSwitchCommand(
            self.api,
            self.router_name,
            self.interconnect_switch_name,
            lrp_ips=_get_gw_ips_for_switch(self.api, self.n_switch),
        ).run_idl(txn)

        ReconcileMainRouterPoliciesForProviderCommand(
            self.api,
            self.interconnect_switch_name,
            related_resource=self.n_switch,
        ).run_idl(txn)


class ReconcileMainRouterPoliciesForProviderCommand(ovs_cmd.BaseCommand):
    def __init__(self, api, interconnect_switch_name, related_resource=None):
        super().__init__(api)
        self.interconnect_switch_name = interconnect_switch_name
        self.related_resource = related_resource

    def run_idl(self, txn):
        router = _get_main_router(self.api)
        lrp_interconnect_name = helpers.get_lrp_name(
            router.name, self.interconnect_switch_name)
        for lrp in helpers.lrps_to_chassis_routers(router):
            ReconcileMainRouterPoliciesCommand(
                self.api,
                router,
                lrp_interconnect_name,
                lrp,
                self.related_resource,
            ).run_idl(txn)


class ReconcileMainRouterPoliciesCommand(ovs_cmd.BaseCommand):
    def __init__(self, api, router, interconnect_lrp_name, chassis_lrp,
                 related_resource=None):
        super().__init__(api)
        self.router = router
        self.interconnect_lrp_name = interconnect_lrp_name
        self.chassis_lrp = chassis_lrp
        self.related_resource = related_resource

    def run_idl(self, txn):
        lrp_peer_ip = _get_lrp_peer_ip(self.api, self.chassis_lrp)
        external_ids = {}

        if self.related_resource:
            external_ids[constants.RELATED_RESOURCE_TAG] = str(
                self.related_resource.uuid)

        # An egress policy to reroute traffic to the chassis router that is
        # local to the chassis where the traffic originated from
        _LrPolicyAddCommand(
            self.api,
            self.router.name,
            priority=10,
            match=f'inport==\"{self.interconnect_lrp_name}\" '
                    f'&& is_chassis_resident(\"cr-{self.chassis_lrp.name}\")',
            action='reroute',
            output_port=self.chassis_lrp,
            nexthops=[lrp_peer_ip],
            external_ids=external_ids,
        ).run_idl(txn)


class ReconcileRouterCommand(_LrAddCommand):
    def __init__(self, api, name):
        # We need to set policies and static_routes to empty list because IDL
        # won't have that set until the transaction is committed
        super().__init__(
            api, name, may_exist=True, policies=[], static_routes=[])

    def run_idl(self, txn):
        super().run_idl(txn)

        for key, value in self.options.items():
            self.row_result.setkey('options', key, value)

    @property
    def options(self):
        return {}


class ReconcileMainRouterCommand(ReconcileRouterCommand):
    def __init__(self, api):
        name = bgp_config.get_main_router_name()
        super().__init__(api, name)

    @property
    def options(self):
        return {
            constants.LR_OPTIONS_DYNAMIC_ROUTING: 'true',
            constants.LR_OPTIONS_DYNAMIC_ROUTING_REDISTRIBUTE:
                constants.BGP_ROUTER_REDISTRIBUTE,
            constants.LR_OPTIONS_DYNAMIC_ROUTING_VRF_ID:
                str(bgp_config.get_main_router_vrf_id()),
        }

    def run_idl(self, txn):
        super().run_idl(txn)

        # Create a fake LRP just to get the fake route in place
        fake_lrp = _run_idl_command(_LrpAddCommand(
            self.api,
            self.router,
            f'{self.router}-dead-lrp',
            may_exist=True,
        ), txn)

        # A fake route to get over the routing stage in routers logical flows
        # This is required for the egress policy to work
        nb_cmd.LrRouteAddCommand(
            self.api,
            self.router,
            '0.0.0.0/0',
            helpers.ipv6_link_local_from_mac(fake_lrp.mac),
            may_exist=True,
        ).run_idl(txn)


class ReconcileChassisRouterCommand(ReconcileRouterCommand):
    def __init__(self, api, chassis):
        self.chassis = chassis
        router_name = helpers.get_chassis_router_name(self.chassis.name)
        super().__init__(api, router_name)

    @property
    def options(self):
        return {
            'chassis': self.chassis.name,
            constants.LR_OPTIONS_DYNAMIC_ROUTING: 'true',
            constants.LR_OPTIONS_DYNAMIC_ROUTING_VRF_ID:
                str(bgp_config.get_chassis_router_vrf_id()),
        }


class ConnectRouterToSwitchCommand(ovs_cmd.BaseCommand):
    def __init__(self, api, router_name, switch_name, lrp_ips=None):
        super().__init__(api)
        self.router_name = router_name
        self.switch_name = switch_name
        self.lrp_name = helpers.get_lrp_name(
            self.router_name, self.switch_name)
        self.lrp_ips = lrp_ips or []

    def run_idl(self, txn):
        _LrpAddCommand(
            self.api,
            self.router_name,
            self.lrp_name,
            networks=self.lrp_ips,
        ).run_idl(txn)

        lsp_name = helpers.get_lsp_name(self.switch_name, self.router_name)
        _LspAddCommand(
            self.api,
            self.switch_name,
            lsp_name,
            addresses=ovn_const.DEFAULT_ADDR_FOR_LSP_WITH_PEER,
            type=ovn_const.LSP_TYPE_ROUTER,
            options={'router-port': self.lrp_name},
        ).run_idl(txn)


class ConnectChassisRouterToSwitchCommand(ConnectRouterToSwitchCommand):
    def __init__(self, api, router_name, switch_name, network_name):
        super().__init__(api, router_name, switch_name)
        self.network_name = network_name

    def run_idl(self, txn):
        super().run_idl(txn)
        ovs_cmd.DbSetCommand(
            self.api,
            'Logical_Router_Port',
            self.lrp_name,
            external_ids={
                constants.LRP_NETWORK_NAME_EXT_ID_KEY: self.network_name,
            },
            options={
                constants.LRP_OPTIONS_PORT_NAME: self.network_name,
            },
        ).run_idl(txn)


class ConnectChassisRouterToMainRouterCommand(ovs_cmd.BaseCommand):
    def __init__(self, api, chassis, hcg_uuid):
        super().__init__(api)
        self.chassis = chassis
        self.hcg_uuid = hcg_uuid
        self.router_name = helpers.get_chassis_router_name(self.chassis.name)

    def validate_prerequisites(self):
        for router_name in [
                bgp_config.get_main_router_name(), self.router_name]:
            try:
                self.api.lookup('Logical_Router', router_name)
            except idlutils.RowNotFound:
                raise exceptions.ReconcileError(
                    f"Router {router_name} not found")

    def run_idl(self, txn):
        self.validate_prerequisites()

        main_router_name = bgp_config.get_main_router_name()

        lrp_main = helpers.get_lrp_name(main_router_name, self.router_name)
        lrp_ch = helpers.get_lrp_name(self.router_name, main_router_name)

        _LrpAddCommand(
            self.api,
            self.router_name,
            lrp_ch,
            peer=lrp_main,
            options={
                constants.LRP_OPTIONS_DYNAMIC_ROUTING_MAINTAIN_VRF: 'true',
            },
        ).run_idl(txn)

        _LrpAddCommand(
            self.api,
            main_router_name,
            lrp_main,
            peer=lrp_ch,
            ha_chassis_group=self.hcg_uuid,
            options={
                constants.LRP_OPTIONS_DYNAMIC_ROUTING_MAINTAIN_VRF: 'true'},
            external_ids={
                constants.BGP_LRP_TO_CHASSIS: self.router_name,
            },
        ).run_idl(txn)


class ReconcileChassisCommand(ovs_cmd.BaseCommand):
    """Reconcile all BGP components for a chassis

    The command reconciles the chassis router and all its configured peer
    connections based on the configured BGP bridges on the given chassis. It
    creates a logical switch with a localnet port connected to the BGP bridge,
    creates routes in and out on the router and connects the router to the main
    router with a peer connection.
    """

    def __init__(self, api, chassis):
        super().__init__(api)
        self.chassis = chassis

    def run_idl(self, txn):
        hcg_name = helpers.get_hcg_name(self.chassis.name)
        hcg = _run_idl_command(_HAChassisGroupAddCommand(
            self.api,
            hcg_name), txn)

        nb_cmd.HAChassisGroupAddChassisCommand(
            self.api,
            hcg.uuid,
            self.chassis.name, constants.HA_CHASSIS_GROUP_PRIORITY
        ).run_idl(txn)

        ReconcileChassisRouterCommand(
            self.api,
            self.chassis,
        ).run_idl(txn)

        # Connect chassis router to the main router
        ConnectChassisRouterToMainRouterCommand(
            self.api,
            self.chassis,
            hcg.uuid,
        ).run_idl(txn)

        for bgp_bridge in helpers.get_chassis_bgp_bridges(self.chassis):
            ReconcileChassisPeerCommand(
                self.api,
                self.chassis,
                network_name=bgp_bridge,
            ).run_idl(txn)


class ReconcileChassisPeerCommand(ovs_cmd.BaseCommand):
    """The command reconciles a BGP peer connection for a chassis

    The BGP peer connection is based on the BGP bridge chassis configuration.
    It creates a logical switch with a localnet port connected to the BGP
    bridge. All traffic is routed out to the localnet port but there is a
    policy based on the inport, so traffic coming from the main BGP router is
    rerouted with ECMP to the peer IP, that typically resides on the
    neighboring physical switch.
    """
    def __init__(
            self, api, chassis, network_name):
        super().__init__(api)
        self.chassis = chassis
        self.network_name = network_name

    @property
    def chassis_router_name(self):
        return helpers.get_chassis_router_name(self.chassis.name)

    @property
    def switch_name(self):
        return helpers.get_chassis_peer_switch_name(
            self.chassis.name, self.network_name)

    def run_idl(self, txn):
        CreateSwitchWithLocalnetCommand(
            self.api,
            self.switch_name,
            self.network_name,
        ).run_idl(txn)

        ConnectChassisRouterToSwitchCommand(
            self.api,
            self.chassis_router_name,
            self.switch_name,
            network_name=self.network_name,
        ).run_idl(txn)

        peer_switch_lrp_name = helpers.get_lrp_name(
            self.chassis_router_name, self.switch_name)
        match=f'inport==\"{peer_switch_lrp_name}\"'
        lrp_to_main_router_name = helpers.get_lrp_name(
            self.chassis_router_name, bgp_config.get_main_router_name())
        lrp = self.api.lookup('Logical_Router_Port', lrp_to_main_router_name)
        lrp_peer_ip = _get_lrp_peer_ip(self.api, lrp)
        _LrPolicyAddCommand(
            self.api,
            self.chassis_router_name,
            priority=10,
            match=match,
            action=ovsdbapp_const.POLICY_ACTION_REROUTE,
            nexthops=[lrp_peer_ip],
            output_port=lrp,
        ).run_idl(txn)


class FullSyncBGPTopologyCommand(ovs_cmd.BaseCommand):
    def __init__(self, nb_api, sb_api):
        super().__init__(nb_api)
        self.sb_api = sb_api

    def run_idl(self, txn):
        LOG.debug("BGP full sync topology started")
        self.reconcile_central(txn)
        self.reconcile_all_chassis(txn)
        self.reconcile_neutron_switches(txn)
        LOG.debug("BGP full sync topology completed")

    def reconcile_all_chassis(self, txn):
        for chassis in self.sb_api.tables['Chassis_Private'].rows.values():
            ReconcileChassisCommand(self.api, chassis).run_idl(txn)

    def reconcile_neutron_switches(self, txn):
        for switch in _get_all_provider_switches(self.api):
            ReconcileNeutronSwitchCommand(
                self.api,
                switch,
            ).run_idl(txn)

    def reconcile_central(self, txn):
        ReconcileMainRouterCommand(
            self.api,
        ).run_idl(txn)



# Copyright 2024 Red Hat, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from unittest import mock

import netaddr
from neutron_lib.api import attributes
from neutron_lib import constants as n_const
from neutron_lib import context
from neutron_lib import exceptions
from neutron_lib.utils import net as net_utils
from oslo_utils import uuidutils

from neutron.extensions import tagging
from neutron.objects import network as network_obj
from neutron.objects import network_segment_range as network_segment_range_obj
from neutron.objects import ports as ports_obj
from neutron.objects.qos import policy as policy_obj
from neutron.objects import router as router_obj
from neutron.objects import securitygroup as securitygroup_obj
from neutron.objects import subnet as subnet_obj
from neutron.objects import subnetpool as subnetpool_obj
from neutron.objects import trunk as trunk_obj
from neutron.tests.unit import testlib_api


class TaggingControllerDbTestCase(testlib_api.WebTestCase):
    def setUp(self):
        super().setUp()
        self.user_id = uuidutils.generate_uuid()
        self.project_id = uuidutils.generate_uuid()
        self.ctx = context.Context(user_id=self.user_id,
                                   project_id=self.project_id,
                                   is_admin=False)
        self.tc = tagging.TaggingController()
        mock.patch.dict(
            attributes.RESOURCES,
            {
                'floatingips': {
                    'id': {'primary_key': True},
                    'router_id': {'required_by_policy': True},
                    'tenant_id': {'required_by_policy': True}
                },
                'network_segment_ranges': {
                    'id': {'primary_key': True},
                    'project_id': {'required_by_policy': True}
                },
                'policies':
                {
                    'id': {'primary_key': True},
                    'tenant_id': {'required_by_policy': True}
                },
                'routers':
                {
                    'id': {'primary_key': True},
                    'tenant_id': {'required_by_policy': True}
                },
                'security_groups':
                {
                    'id': {'primary_key': True},
                    'tenant_id': {'required_by_policy': True}
                },
                'trunks':
                {
                    'id': {'primary_key': True},
                    'port_id': {'required_by_policy': True},
                    'tenant_id': {'required_by_policy': True}
                }
            }
        ).start()

    def test_all_ovo_cls_have_a_reference(self):
        tc_supported_resources = set(self.tc.supported_resources.keys())
        ovo_resources = set(tagging.OVO_CLS.keys())
        self.assertEqual(tc_supported_resources, ovo_resources)

    def _check_resource_info(self, obj, obj_type):
        id_key = self.tc.supported_resources[obj_type] + '_id'
        res = self.tc._get_resource_info(self.ctx, {id_key: obj['id']})
        reference = tagging.ResourceInfo(self.project_id, obj_type, obj)
        self.assertEqual(reference, res)

    def test__get_resource_info_floatingips(self):
        ext_net_id = uuidutils.generate_uuid()
        fip_port_id = uuidutils.generate_uuid()
        fip_id = uuidutils.generate_uuid()
        network_obj.Network(
            self.ctx, id=ext_net_id, project_id=self.project_id).create()
        network_obj.ExternalNetwork(
            self.ctx, project_id=self.project_id,
            network_id=ext_net_id).create()
        mac_str = next(net_utils.random_mac_generator(
            ['ca', 'fe', 'ca', 'fe']))
        mac = netaddr.EUI(mac_str)
        ports_obj.Port(
            self.ctx, id=fip_port_id, project_id=self.project_id,
            mac_address=mac, network_id=ext_net_id, admin_state_up=True,
            status='UP', device_id='', device_owner='').create()
        ip_address = netaddr.IPAddress('1.2.3.4')
        router_obj.FloatingIP(
            self.ctx, id=fip_id, project_id=self.project_id,
            floating_network_id=ext_net_id, floating_port_id=fip_port_id,
            floating_ip_address=ip_address).create()
        expected_fip = {
            'attributes_to_update': ['tags'],
            'id': fip_id,
            'tenant_id': self.project_id,
            'project_id': self.project_id
        }
        self._check_resource_info(expected_fip, 'floatingips')

    def test__get_resource_info_network_segment_ranges(self):
        srange_id = uuidutils.generate_uuid()
        network_segment_range_obj.NetworkSegmentRange(
            self.ctx, id=srange_id, project_id=self.project_id,
            shared=False, network_type=n_const.TYPE_GENEVE,
            minimum=1, maximum=100).create()
        expected_segment = {
            'attributes_to_update': ['tags'],
            'id': srange_id,
            'project_id': self.project_id
        }
        self._check_resource_info(expected_segment, 'network_segment_ranges')

    def test__get_resource_info_networks(self):
        net_id = uuidutils.generate_uuid()
        network_obj.Network(
            self.ctx, id=net_id, project_id=self.project_id).create()
        expected_net = {
            'attributes_to_update': ['tags'],
            'id': net_id,
            'tenant_id': self.project_id,
            'project_id': self.project_id,
            'shared': False,
        }
        self._check_resource_info(expected_net, 'networks')

    def test__get_resource_info_policies(self):
        qos_id = uuidutils.generate_uuid()
        policy_obj.QosPolicy(
            self.ctx, id=qos_id, project_id=self.project_id).create()
        expected_qos = {
            'attributes_to_update': ['tags'],
            'id': qos_id,
            'tenant_id': self.project_id,
            'project_id': self.project_id
        }
        self._check_resource_info(expected_qos, 'policies')

    def test__get_resource_info_ports(self):
        net_id = uuidutils.generate_uuid()
        port_id = uuidutils.generate_uuid()
        network_obj.Network(
            self.ctx, id=net_id, project_id=self.project_id).create()
        mac_str = next(net_utils.random_mac_generator(
            ['ca', 'fe', 'ca', 'fe']))
        mac = netaddr.EUI(mac_str)
        ports_obj.Port(
            self.ctx, id=port_id, project_id=self.project_id,
            mac_address=mac, network_id=net_id, admin_state_up=True,
            status='UP', device_id='', device_owner='').create()
        expected_port = {
            'attributes_to_update': ['tags'],
            'id': port_id,
            'tenant_id': self.project_id,
            'project_id': self.project_id,
            'network_id': net_id,
            'status': 'UP',
        }
        self._check_resource_info(expected_port, 'ports')

    def test__get_resource_info_routers(self):
        router_id = uuidutils.generate_uuid()
        router_obj.Router(
            self.ctx, id=router_id, project_id=self.project_id).create()
        expected_router = {
            'attributes_to_update': ['tags'],
            'id': router_id,
            'tenant_id': self.project_id,
            'project_id': self.project_id
        }
        self._check_resource_info(expected_router, 'routers')

    def test__get_resource_info_security_groups(self):
        sg_id = uuidutils.generate_uuid()
        securitygroup_obj.SecurityGroup(
            self.ctx, id=sg_id, project_id=self.project_id,
            is_default=True).create()
        expected_sg = {
            'attributes_to_update': ['tags'],
            'id': sg_id,
            'tenant_id': self.project_id,
            'project_id': self.project_id
        }
        self._check_resource_info(expected_sg, 'security_groups')

    def test__get_resource_info_subnets(self):
        net_id = uuidutils.generate_uuid()
        subnet_id = uuidutils.generate_uuid()
        network_obj.Network(
            self.ctx, id=net_id, project_id=self.project_id).create()
        cidr = netaddr.IPNetwork('1.2.3.0/24')
        subnet_obj.Subnet(
            self.ctx, id=subnet_id, project_id=self.project_id,
            ip_version=n_const.IP_VERSION_4, cidr=cidr,
            network_id=net_id).create()
        expected_subnet = {
            'attributes_to_update': ['tags'],
            'id': subnet_id,
            'ip_version': n_const.IP_VERSION_4,
            'shared': False,
            'network_id': net_id,
            'tenant_id': self.project_id,
            'project_id': self.project_id
        }
        self._check_resource_info(expected_subnet, 'subnets')

    def test__get_resource_info_subnetpools(self):
        sp_id = uuidutils.generate_uuid()
        subnetpool_obj.SubnetPool(
            self.ctx, id=sp_id, project_id=self.project_id,
            ip_version=n_const.IP_VERSION_4, default_prefixlen=26,
            min_prefixlen=28, max_prefixlen=26).create()
        expected_sp = {
            'attributes_to_update': ['tags'],
            'id': sp_id,
            'tenant_id': self.project_id,
            'project_id': self.project_id,
            'ip_version': n_const.IP_VERSION_4,
            'shared': False,
            'is_default': False,
            'prefixes': [],
        }
        self._check_resource_info(expected_sp, 'subnetpools')

    def test__get_resource_info_trunks(self):
        trunk_id = uuidutils.generate_uuid()
        net_id = uuidutils.generate_uuid()
        port_id = uuidutils.generate_uuid()
        network_obj.Network(
            self.ctx, id=net_id, project_id=self.project_id).create()
        mac_str = next(net_utils.random_mac_generator(
            ['ca', 'fe', 'ca', 'fe']))
        mac = netaddr.EUI(mac_str)
        ports_obj.Port(
            self.ctx, id=port_id, project_id=self.project_id,
            mac_address=mac, network_id=net_id, admin_state_up=True,
            status='UP', device_id='', device_owner='').create()
        trunk_obj.Trunk(
            self.ctx, id=trunk_id, project_id=self.project_id,
            port_id=port_id).create()
        expected_trunk = {
            'attributes_to_update': ['tags'],
            'id': trunk_id,
            'tenant_id': self.project_id,
            'project_id': self.project_id,
            'port_id': port_id
        }
        self._check_resource_info(expected_trunk, 'trunks')

    def test__get_resource_info_object_not_present(self):
        missing_id = uuidutils.generate_uuid()
        p_id = self.tc.supported_resources['trunks'] + '_id'
        res = self.tc._get_resource_info(self.ctx, {p_id: missing_id})
        self.assertEqual(tagging.EMPTY_RESOURCE_INFO, res)

    def test__get_resource_info_wrong_resource(self):
        missing_id = uuidutils.generate_uuid()
        res = self.tc._get_resource_info(self.ctx,
                                         {'wrong_resource_id': missing_id})
        self.assertEqual(tagging.EMPTY_RESOURCE_INFO, res)

    def test_create_tags_for_resource_below_max_tags_limit(self):
        req = mock.Mock(context=self.ctx)
        tags = ['tag-%i' % i for i in range(tagging.MAX_TAGS_COUNT - 1)]
        body = {'tags': tags}
        obj_mock = {
            'id': uuidutils.generate_uuid()
        }
        with mock.patch.object(self.tc, '_get_resource_info') as get_res, \
                mock.patch('neutron.policy.enforce') as policy_enforce, \
                mock.patch.object(tagging, 'notify_tag_action') as notify:
            self.tc.plugin = mock.Mock()
            get_res.return_value = mock.MagicMock(obj_type='networks',
                                                  obj=obj_mock)

            self.tc.create(req, body)

            get_res.assert_called_once_with(self.ctx, mock.ANY, tags=tags)
            policy_enforce.assert_called_once_with(
                self.ctx, mock.ANY, obj_mock)
            self.tc.plugin.create_tags.assert_called_once_with(
                self.ctx, mock.ANY, mock.ANY, body)
            notify.assert_has_calls([
                mock.call(self.ctx, 'create.start', mock.ANY, mock.ANY, tags),
                mock.call(self.ctx, 'create.end', mock.ANY, mock.ANY, tags)])

    def test_create_tags_for_resource_over_max_tags_limit(self):
        req = mock.Mock(context=self.ctx)
        tags = ['tag-%i' % i for i in range(tagging.MAX_TAGS_COUNT + 1)]
        body = {'tags': tags}
        obj_mock = {
            'id': uuidutils.generate_uuid()
        }
        with mock.patch.object(self.tc, '_get_resource_info') as get_res, \
                mock.patch('neutron.policy.enforce') as policy_enforce, \
                mock.patch.object(tagging, 'notify_tag_action') as notify:
            self.tc.plugin = mock.Mock()
            get_res.return_value = mock.MagicMock(obj_type='networks',
                                                  obj=obj_mock)

            self.assertRaises(
                exceptions.BadRequest,
                self.tc.create, req, body)

            get_res.assert_called_once_with(self.ctx, mock.ANY, tags=tags)
            policy_enforce.assert_called_once_with(
                self.ctx, mock.ANY, obj_mock)
            self.tc.plugin.create_tags.assert_not_called()
            notify.assert_not_called()

    def test_update_tag_for_resource_below_max_tags_limit(self):
        req = mock.Mock(context=self.ctx)
        kwargs = {'network_id': uuidutils.generate_uuid()}
        existing_tags = [
            'tag-%i' % i for i in range(tagging.MAX_TAGS_COUNT - 2)]
        new_tag = 'new-tag'
        obj_mock = {
            'id': uuidutils.generate_uuid()
        }
        with mock.patch.object(self.tc, '_get_resource_info') as get_res, \
                mock.patch('neutron.policy.enforce') as policy_enforce, \
                mock.patch.object(tagging, 'notify_tag_action') as notify:
            self.tc.plugin = mock.Mock()
            self.tc.plugin.get_tags.return_value = {'tags': existing_tags}
            get_res.return_value = mock.MagicMock(obj_type='networks',
                                                  obj=obj_mock)

            self.tc.update(req, new_tag, **kwargs)

            get_res.assert_called_once_with(self.ctx, kwargs, tags=[new_tag])
            policy_enforce.assert_called_once_with(
                self.ctx, mock.ANY, obj_mock)
            self.tc.plugin.get_tags.assert_called_once_with(
                self.ctx, mock.ANY, mock.ANY)
            self.tc.plugin.update_tag.assert_called_once_with(
                self.ctx, mock.ANY, mock.ANY, new_tag)
            notify.assert_has_calls([
                mock.call(
                    self.ctx, 'create.start', mock.ANY, mock.ANY, [new_tag]),
                mock.call(
                    self.ctx, 'create.end', mock.ANY, mock.ANY, [new_tag])])

    def test_update_exising_tag_for_resource_above_max_tags_limit(self):
        """Test to ensure that TaggingController.update() method is idempotent.
        """

        req = mock.Mock(context=self.ctx)
        kwargs = {'network_id': uuidutils.generate_uuid()}
        existing_tags = [
            'tag-%i' % i for i in range(tagging.MAX_TAGS_COUNT)]
        new_tag = existing_tags[0]
        obj_mock = {
            'id': uuidutils.generate_uuid()
        }
        with mock.patch.object(self.tc, '_get_resource_info') as get_res, \
                mock.patch('neutron.policy.enforce') as policy_enforce, \
                mock.patch.object(tagging, 'notify_tag_action') as notify:
            self.tc.plugin = mock.Mock()
            self.tc.plugin.get_tags.return_value = {'tags': existing_tags}
            get_res.return_value = mock.MagicMock(obj_type='networks',
                                                  obj=obj_mock)

            self.tc.update(req, new_tag, **kwargs)

            get_res.assert_called_once_with(self.ctx, kwargs, tags=[new_tag])
            policy_enforce.assert_called_once_with(
                self.ctx, mock.ANY, obj_mock)
            self.tc.plugin.get_tags.assert_called_once_with(
                self.ctx, mock.ANY, mock.ANY)
            self.tc.plugin.update_tag.assert_called_once_with(
                self.ctx, mock.ANY, mock.ANY, new_tag)
            notify.assert_has_calls([
                mock.call(
                    self.ctx, 'create.start', mock.ANY, mock.ANY, [new_tag]),
                mock.call(
                    self.ctx, 'create.end', mock.ANY, mock.ANY, [new_tag])])

    def test_update_tag_for_resource_over_max_tags_limit(self):
        req = mock.Mock(context=self.ctx)
        kwargs = {'network_id': uuidutils.generate_uuid()}
        existing_tags = ['tag-%i' % i for i in range(tagging.MAX_TAGS_COUNT)]
        new_tag = 'new-tag'
        obj_mock = {
            'id': uuidutils.generate_uuid()
        }
        with mock.patch.object(self.tc, '_get_resource_info') as get_res, \
                mock.patch('neutron.policy.enforce') as policy_enforce, \
                mock.patch.object(tagging, 'notify_tag_action') as notify:
            self.tc.plugin = mock.Mock()
            self.tc.plugin.get_tags.return_value = {'tags': existing_tags}
            get_res.return_value = mock.MagicMock(obj_type='networks',
                                                  obj=obj_mock)

            self.assertRaises(
                exceptions.BadRequest,
                self.tc.update, req, new_tag, **kwargs)

            get_res.assert_called_once_with(self.ctx, kwargs, tags=[new_tag])
            policy_enforce.assert_called_once_with(
                self.ctx, mock.ANY, obj_mock)
            self.tc.plugin.get_tags.assert_called_once_with(
                self.ctx, mock.ANY, mock.ANY)
            self.tc.plugin.update_tag.assert_not_called()
            notify.assert_not_called()

    def test_update_all_tags_for_resource_below_max_tags_limit(self):
        req = mock.Mock(context=self.ctx)
        tags = ['tag-%i' % i for i in range(tagging.MAX_TAGS_COUNT - 1)]
        body = {'tags': tags}
        obj_mock = {
            'id': uuidutils.generate_uuid()
        }
        with mock.patch.object(self.tc, '_get_resource_info') as get_res, \
                mock.patch('neutron.policy.enforce') as policy_enforce, \
                mock.patch.object(tagging, 'notify_tag_action') as notify:
            self.tc.plugin = mock.Mock()
            get_res.return_value = mock.MagicMock(obj_type='networks',
                                                  obj=obj_mock)

            self.tc.update_all(req, body)

            get_res.assert_called_once_with(self.ctx, mock.ANY, tags=tags)
            policy_enforce.assert_called_once_with(
                self.ctx, mock.ANY, obj_mock)
            self.tc.plugin.update_tags.assert_called_once_with(
                self.ctx, mock.ANY, mock.ANY, body)
            notify.assert_has_calls([
                mock.call(self.ctx, 'update.start', mock.ANY, mock.ANY, tags),
                mock.call(self.ctx, 'update.end', mock.ANY, mock.ANY, tags)])

    def test_update_all_tags_for_resource_over_max_tags_limit(self):
        req = mock.Mock(context=self.ctx)
        tags = ['tag-%i' % i for i in range(tagging.MAX_TAGS_COUNT + 1)]
        body = {'tags': tags}
        obj_mock = {
            'id': uuidutils.generate_uuid()
        }
        with mock.patch.object(self.tc, '_get_resource_info') as get_res, \
                mock.patch('neutron.policy.enforce') as policy_enforce, \
                mock.patch.object(tagging, 'notify_tag_action') as notify:
            self.tc.plugin = mock.Mock()
            get_res.return_value = mock.MagicMock(obj_type='networks',
                                                  obj=obj_mock)

            self.assertRaises(
                exceptions.BadRequest,
                self.tc.update_all, req, body)

            get_res.assert_called_once_with(self.ctx, mock.ANY, tags=tags)
            policy_enforce.assert_called_once_with(
                self.ctx, mock.ANY, obj_mock)
            self.tc.plugin.update_tags.assert_not_called()
            notify.assert_not_called()



# Copyright 2025 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from alembic import op
from oslo_log import log as logging
from sqlalchemy import exc

from neutron.db import migration


LOG = logging.getLogger(__name__)


# Make "tag" resources case sensitive, using collate/charset
# "utf8mb4_bin/utf8mb4"
#
# Revision ID: d553edeb540f
# Revises: ad80a9f07c5c
# Create Date: 2025-06-17 23:13:24.272747

# revision identifiers, used by Alembic.
revision = 'd553edeb540f'
down_revision = 'ad80a9f07c5c'

# milestone identifier, used by neutron-db-manage
neutron_milestone = [migration.RELEASE_2025_2]


def upgrade():
    try:
        op.execute('ALTER TABLE tags CONVERT TO CHARACTER SET utf8mb4 '
                   'COLLATE utf8mb4_bin')
        op.execute('ALTER TABLE tags CHARACTER SET utf8mb4 '
                   'COLLATE utf8mb4_bin')
    except exc.OperationalError as _exc:
        if 'Unknown collation' in str(_exc):
            LOG.error('Collation "utf8mb4_bin" does not exist; the Neutron '
                      '"tag" table will remain case insensitive.')
        else:
            raise _exc



# Copyright 2025 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron_lib.api.definitions import external_net
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory

from neutron.common.ovn import utils as ovn_utils
from neutron.common import utils as n_utils
from neutron.services.bgp import constants as bgp_constants
from neutron.tests.functional import base
from neutron.tests.unit.api import test_extensions
from neutron.tests.unit.extensions import test_l3


class TestLogicalRouterPortEvent(
    base.TestOVNFunctionalBase,
    test_l3.L3NatTestCaseMixin):

    def setUp(self, **kwargs):
        super().setUp(**kwargs)
        self.chassis = self.add_fake_chassis('ovs-host1')
        self.l3_plugin = directory.get_plugin(plugin_constants.L3)
        self.l3_plugin._post_fork_initialize(mock.ANY, mock.ANY, mock.ANY)
        self.ext_api = test_extensions.setup_extensions_middleware(
            test_l3.L3TestExtensionManager())
        kwargs = {'arg_list': (external_net.EXTERNAL,),
                  external_net.EXTERNAL: True}
        self.net_ext = self._make_network(
            self.fmt, 'net_ext', True, as_admin=True, **kwargs)
        self.subnet = self._make_subnet(self.fmt, self.net_ext, '20.0.10.1',
                                        '20.0.10.0/24')
        self.router = self._make_router(self.fmt, self._project_id)
        self.router_id = self.router['router']['id']
        self.net_ext_id = self.net_ext['network']['id']
        self.subnet_id = self.subnet['subnet']['id']

    def test_add_and_delete_gw_network(self):
        def is_called():
            try:
                mock_update_router.assert_called_once_with(
                    mock.ANY, self.router_id)
                return True
            except AssertionError:
                return False

        with mock.patch.object(
                self.l3_plugin._ovn_client,
                'update_router_ha_chassis_group') as mock_update_router:
            self._add_external_gateway_to_router(self.router_id,
                                                 self.net_ext_id)
            n_utils.wait_until_true(is_called, timeout=10)
            mock_update_router.reset_mock()
            self._remove_external_gateway_from_router(
                self.router_id, self.net_ext_id, external_gw_info={})
            n_utils.wait_until_true(is_called, timeout=10)

    def test_add_private_network(self):
        def is_called():
            try:
                mock_link.assert_called_once_with(
                    mock.ANY, self.net_ext_id, self.router_id)
                return True
            except AssertionError:
                return False

        with mock.patch.object(
                self.l3_plugin._ovn_client,
                'link_network_ha_chassis_group') as mock_link:
            self._router_interface_action(
                'add', self.router_id, self.subnet_id, None)
            n_utils.wait_until_true(is_called, timeout=10)

    def test_delete_private_network(self):
        def is_called():
            try:
                mock_unlink.assert_called_once_with(self.net_ext_id)
                return True
            except AssertionError:
                return False

        with mock.patch.object(
                self.l3_plugin._ovn_client,
                'link_network_ha_chassis_group'), \
                mock.patch.object(
                    self.l3_plugin._ovn_client,
                    'unlink_network_ha_chassis_group') as mock_unlink:
            self._router_interface_action(
                'add', self.router_id, self.subnet_id, None)
            self._router_interface_action(
                'remove', self.router_id, self.subnet_id, None)
            n_utils.wait_until_true(is_called, timeout=10)

    def test_delete_router(self):
        # The ``Logical_Router`` deletion triggers the
        # ``LogicalRouterPortEvent`` event, but nothing is executed/called.
        def is_called():
            try:
                mock_update_router.assert_called_once_with(
                    mock.ANY, self.router_id)
                return True
            except AssertionError:
                return False

        with mock.patch.object(
                self.l3_plugin._ovn_client,
                'update_router_ha_chassis_group') as mock_update_router:
            self._add_external_gateway_to_router(self.router_id,
                                                 self.net_ext_id)
            n_utils.wait_until_true(is_called, timeout=10)
            mock_update_router.reset_mock()
            req = self.new_delete_request('routers', self.router_id)
            req.get_response(self.api)
            self.assertRaises(n_utils.WaitTimeout, n_utils.wait_until_true,
                              is_called, timeout=5)

    def test_add_lrp_not_in_the_neutron_db(self):
        def mock_link_called():
            try:
                mock_link.assert_called_once_with(
                    mock.ANY, self.net_ext_id, self.router_id)
                return True
            except AssertionError:
                return False

        def mock_update_router_called():
            try:
                mock_update_router.assert_called_once_with(
                    mock.ANY, self.router_id)
                return True
            except AssertionError:
                return False

        ovn_client = self.l3_plugin._ovn_client
        with mock.patch.object(
                ovn_client, 'link_network_ha_chassis_group') as mock_link,\
                mock.patch.object(ovn_client,
                'update_router_ha_chassis_group') as mock_update_router:
            ext_ids = {bgp_constants.LRP_NETWORK_NAME_EXT_ID_KEY: 'br-bgp1'}
            self.nb_api.lrp_add(
                ovn_utils.ovn_name(self.router_id), 'ovn-port-bgp',
                mac='02:00:00:00:00:00', networks=['192.168.100.2/30'],
                external_ids=ext_ids).execute(check_error=True)
            self.assertRaises(n_utils.WaitTimeout, n_utils.wait_until_true,
                              mock_link_called, timeout=5)
            self.assertRaises(n_utils.WaitTimeout, n_utils.wait_until_true,
                              mock_update_router_called, timeout=2)

    def test_remove_lrp_not_in_the_neutron_db(self):
        def mock_unlink_called():
            try:
                mock_unlink.assert_called_once_with(
                    mock.ANY, self.net_ext_id, self.router_id)
                return True
            except AssertionError:
                return False

        def mock_update_router_called():
            try:
                mock_update_router.assert_called_once_with(
                    mock.ANY, self.router_id)
                return True
            except AssertionError:
                return False

        ovn_client = self.l3_plugin._ovn_client
        with mock.patch.object(
                ovn_client, 'unlink_network_ha_chassis_group') as mock_unlink,\
                mock.patch.object(ovn_client,
                'update_router_ha_chassis_group') as mock_update_router:
            ext_ids = {bgp_constants.LRP_NETWORK_NAME_EXT_ID_KEY: 'br-bgp1'}
            self.nb_api.lrp_add(
                ovn_utils.ovn_name(self.router_id), 'ovn-port-bgp',
                mac='02:00:00:00:00:00', networks=['192.168.100.2/30'],
                external_ids=ext_ids).execute(check_error=True)
            self.nb_api.lrp_del('ovn-port-bgp').execute(check_error=True)
            self.assertRaises(n_utils.WaitTimeout, n_utils.wait_until_true,
                              mock_unlink_called, timeout=5)
            self.assertRaises(n_utils.WaitTimeout, n_utils.wait_until_true,
                              mock_update_router_called, timeout=2)


class TestLogicalRouterPortGatewayChassisEvent(
    base.TestOVNFunctionalBase,
    test_l3.L3NatTestCaseMixin):

    def setUp(self, **kwargs):
        super().setUp(**kwargs)
        self.chassis = self.add_fake_chassis('ovs-host1')
        self.l3_plugin = directory.get_plugin(plugin_constants.L3)
        self.l3_plugin._post_fork_initialize(mock.ANY, mock.ANY, mock.ANY)
        self.ext_api = test_extensions.setup_extensions_middleware(
            test_l3.L3TestExtensionManager())
        kwargs = {'arg_list': (external_net.EXTERNAL,),
                  external_net.EXTERNAL: True}
        self.net_ext = self._make_network(
            self.fmt, 'net_ext', True, as_admin=True, **kwargs)
        self.subnet = self._make_subnet(self.fmt, self.net_ext, '20.0.10.1',
                                        '20.0.10.0/24')
        self.router = self._make_router(self.fmt, self._project_id)
        self.router_id = self.router['router']['id']
        self.net_ext_id = self.net_ext['network']['id']
        self.subnet_id = self.subnet['subnet']['id']

    def test_add_and_remove_gateway_chassis(self):
        def is_called():
            try:
                mock_update_router.assert_called_once_with(
                    mock.ANY, self.router_id)
                return True
            except AssertionError:
                return False

        ch_list = []
        for idx in range(5):
            ch_list.append(self.add_fake_chassis(f'host-{idx}'))
        self._add_external_gateway_to_router(self.router_id, self.net_ext_id)
        lr = self.l3_plugin._nb_ovn.lookup('Logical_Router',
                                           ovn_utils.ovn_name(self.router_id))
        lrp_gw = lr.ports[0]
        with mock.patch.object(
                self.l3_plugin._ovn_client,
                'update_router_ha_chassis_group') as mock_update_router:
            for ch_name in ch_list:
                self.l3_plugin._nb_ovn.lrp_set_gateway_chassis(
                    lrp_gw.uuid, ch_name).execute(check_error=True)
                n_utils.wait_until_true(is_called, timeout=10)
                mock_update_router.reset_mock()

            for ch_name in ch_list:
                self.l3_plugin._nb_ovn.lrp_del_gateway_chassis(
                    lrp_gw.uuid, ch_name).execute(check_error=True)
                n_utils.wait_until_true(is_called, timeout=10)
                mock_update_router.reset_mock()



# Copyright (c) 2016 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from neutron import opts
from neutron.tests import base


class OptsTestCase(base.BaseTestCase):

    def test_list_sriov_agent_opts(self):
        sriov_agent_opts = opts.list_sriov_agent_opts()
        self.assertEqual('DEFAULT', sriov_agent_opts[0][0])
        self.assertEqual('sriov_nic', sriov_agent_opts[1][0])
        self.assertEqual('agent', sriov_agent_opts[2][0])



# Copyright (c) 2012 OpenStack Foundation.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from concurrent import futures
import os.path
import queue
import random
import re
import sys
import threading
import time
from unittest import mock

import ddt
import netaddr
from neutron_lib import constants
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import timeutils
from osprofiler import profiler
import testtools

from neutron.common import utils
from neutron.tests import base
from neutron.tests.unit import tests


class _PortRange:
    """A linked list of port ranges."""
    def __init__(self, base, prev_ref=None):
        self.base = base
        self.mask = 0xffff
        self.prev_ref = prev_ref

    @property
    def possible_mask_base(self):
        return self.base & (self.mask << 1)

    @property
    def can_merge(self):
        return (self.prev_ref and
                self.possible_mask_base == self.prev_ref.possible_mask_base and
                self.mask == self.prev_ref.mask)

    def shake(self):
        """Try to merge ranges created earlier.

        If previous number in a list can be merged with current item under
        common mask, it's merged. Then it continues to do the same with the
        rest of the list.
        """
        while self.can_merge:
            self.mask <<= 1
            self.base = self.prev_ref.base
            if self.prev_ref:
                self.prev_ref = self.prev_ref.prev_ref

    def __str__(self):
        return _hex_format(self.base, self.mask)

    def get_list(self):
        if self.prev_ref:
            return self.prev_ref.get_list() + [str(self)]
        return [str(self)]


def _hex_str(num):
    return format(num, '#06x')


def _hex_format(port, mask):
    if mask != 0xffff:
        return f"{_hex_str(port)}/{_hex_str(0xffff & mask)}"
    return _hex_str(port)


def _port_rule_masking(port_min, port_max):
    current = None
    for num in range(port_min, port_max + 1):
        port_range = _PortRange(num, prev_ref=current)
        port_range.shake()
        current = port_range
    return current.get_list()


class TestExceptionLogger(base.BaseTestCase):
    def test_normal_call(self):
        result = "Result"

        @utils.exception_logger()
        def func():
            return result

        self.assertEqual(result, func())

    def test_raise(self):
        result = "Result"

        @utils.exception_logger()
        def func():
            raise RuntimeError(result)

        self.assertRaises(RuntimeError, func)

    def test_spawn_normal(self):
        result = "Result"
        logger = mock.Mock()

        @utils.exception_logger(logger=logger)
        def func(ret):
            ret.append(result)

        ret_value = []
        _thread = threading.Thread(target=func, args=(ret_value, ))
        _thread.start()
        _thread.join()
        self.assertEqual(result, ret_value[0])
        self.assertFalse(logger.called)

    def test_spawn_raise(self):
        result = "Result"
        logger = mock.Mock()

        @utils.exception_logger(logger=logger)
        def func():
            raise RuntimeError(result)

        _thread = threading.Thread(target=func)
        _thread.start()
        self.assertTrue(logger.called)

    def test_pool_spawn_normal(self):
        logger = mock.Mock()
        calls = mock.Mock()

        @utils.exception_logger(logger=logger)
        def func(i):
            calls(i)

        with futures.ThreadPoolExecutor(max_workers=4) as executor:
            fs = [executor.submit(func, i) for i in range(4)]
            all(f.done() for f in fs)

        calls.assert_has_calls([mock.call(0), mock.call(1),
                                mock.call(2), mock.call(3)],
                               any_order=True)
        self.assertFalse(logger.called)

    def test_pool_spawn_raise(self):
        logger = mock.Mock()
        calls = mock.Mock()

        @utils.exception_logger(logger=logger)
        def func(i):
            if i == 2:
                raise RuntimeError(2)
            calls(i)

        with futures.ThreadPoolExecutor(max_workers=4) as executor:
            fs = [executor.submit(func, i) for i in range(4)]
            all(f.done() for f in fs)

        calls.assert_has_calls([mock.call(0), mock.call(1), mock.call(3)],
                               any_order=True)
        self.assertTrue(logger.called)

    def test_wait_until_true(self):

        class FalseException(Exception):
            pass

        self.assertRaises(
            FalseException,
            utils.wait_until_true,
            lambda: False, timeout=1, exception=FalseException)


class TestDvrServices(base.BaseTestCase):

    def _test_is_dvr_serviced(self, device_owner, expected):
        self.assertEqual(expected, utils.is_dvr_serviced(device_owner))

    def test_is_dvr_serviced_with_dhcp_port(self):
        self._test_is_dvr_serviced(constants.DEVICE_OWNER_DHCP, True)

    def test_is_dvr_serviced_with_vm_port(self):
        self._test_is_dvr_serviced(constants.DEVICE_OWNER_COMPUTE_PREFIX, True)


class TestFipServices(base.BaseTestCase):

    def _test_is_fip_serviced(self, device_owner, expected):
        self.assertEqual(expected, utils.is_fip_serviced(device_owner))

    def test_is_fip_serviced_with_dhcp_port(self):
        self._test_is_fip_serviced(constants.DEVICE_OWNER_DHCP, False)

    def test_is_fip_serviced_with_vm_port(self):
        self._test_is_fip_serviced(constants.DEVICE_OWNER_COMPUTE_PREFIX, True)


class TestIpToCidr(base.BaseTestCase):
    def test_ip_to_cidr_ipv4_default(self):
        self.assertEqual('15.1.2.3/32', utils.ip_to_cidr('15.1.2.3'))

    def test_ip_to_cidr_ipv4_prefix(self):
        self.assertEqual('15.1.2.3/24', utils.ip_to_cidr('15.1.2.3', 24))

    def test_ip_to_cidr_ipv4_netaddr(self):
        ip_address = netaddr.IPAddress('15.1.2.3')
        self.assertEqual('15.1.2.3/32', utils.ip_to_cidr(ip_address))

    def test_ip_to_cidr_ipv4_bad_prefix(self):
        self.assertRaises(netaddr.core.AddrFormatError,
                          utils.ip_to_cidr, '15.1.2.3', 33)

    def test_ip_to_cidr_ipv6_default(self):
        self.assertEqual('::1/128', utils.ip_to_cidr('::1'))

    def test_ip_to_cidr_ipv6_prefix(self):
        self.assertEqual('::1/64', utils.ip_to_cidr('::1', 64))

    def test_ip_to_cidr_ipv6_bad_prefix(self):
        self.assertRaises(netaddr.core.AddrFormatError,
                          utils.ip_to_cidr, '2000::1', 129)


class TestCidrIsHost(base.BaseTestCase):
    def test_is_cidr_host_ipv4(self):
        self.assertTrue(utils.is_cidr_host('15.1.2.3/32'))

    def test_is_cidr_host_ipv4_not_cidr(self):
        self.assertRaises(ValueError,
                          utils.is_cidr_host,
                          '15.1.2.3')

    def test_is_cidr_host_ipv6(self):
        self.assertTrue(utils.is_cidr_host('2000::1/128'))

    def test_is_cidr_host_ipv6_netaddr(self):
        net = netaddr.IPNetwork("2000::1")
        self.assertTrue(utils.is_cidr_host(net))

    def test_is_cidr_host_ipv6_32(self):
        self.assertFalse(utils.is_cidr_host('2000::1/32'))

    def test_is_cidr_host_ipv6_not_cidr(self):
        self.assertRaises(ValueError,
                          utils.is_cidr_host,
                          '2000::1')

    def test_is_cidr_host_ipv6_not_cidr_netaddr(self):
        ip_address = netaddr.IPAddress("2000::3")
        self.assertRaises(ValueError,
                          utils.is_cidr_host,
                          ip_address)


class TestIpVersionFromInt(base.BaseTestCase):
    def test_ip_version_from_int_ipv4(self):
        self.assertEqual(constants.IPv4,
                         utils.ip_version_from_int(constants.IP_VERSION_4))

    def test_ip_version_from_int_ipv6(self):
        self.assertEqual(constants.IPv6,
                         utils.ip_version_from_int(constants.IP_VERSION_6))

    def test_ip_version_from_int_illegal_int(self):
        self.assertRaises(ValueError,
                          utils.ip_version_from_int,
                          8)


class TestIsVersionGreaterEqual(base.BaseTestCase):
    def test_is_version_greater_equal_greater(self):
        self.assertTrue(utils.is_version_greater_equal('1.6.2', '1.6.0'))

    def test_is_version_greater_equal_equal(self):
        self.assertTrue(utils.is_version_greater_equal('1.6.2', '1.6.2'))

    def test_is_version_greater_equal_less(self):
        self.assertFalse(utils.is_version_greater_equal('1.6.0', '1.6.2'))


class TestDelayedStringRenderer(base.BaseTestCase):
    def test_call_deferred_until_str(self):
        my_func = mock.MagicMock(return_value='Brie cheese!')
        delayed = utils.DelayedStringRenderer(my_func, 1, 2, key_arg=44)
        self.assertFalse(my_func.called)
        string = "Type: %s" % delayed
        my_func.assert_called_once_with(1, 2, key_arg=44)
        self.assertEqual("Type: Brie cheese!", string)

    def test_not_called_with_low_log_level(self):
        LOG = logging.getLogger(__name__)
        # make sure we return logging to previous level
        current_log_level = LOG.logger.getEffectiveLevel()
        self.addCleanup(LOG.logger.setLevel, current_log_level)

        my_func = mock.MagicMock()
        delayed = utils.DelayedStringRenderer(my_func)

        # set to warning so we shouldn't be logging debug messages
        LOG.logger.setLevel(logging.logging.WARNING)
        LOG.debug("Hello %s", delayed)
        self.assertFalse(my_func.called)

        # but it should be called with the debug level
        LOG.logger.setLevel(logging.logging.DEBUG)
        LOG.debug("Hello %s", delayed)
        self.assertTrue(my_func.called)


class TestPortRuleMasking(base.BaseTestCase):
    def test_port_rule_wrong_input(self):
        with testtools.ExpectedException(ValueError):
            utils.port_rule_masking(12, 5)

    def compare_port_ranges_results(self, port_min, port_max):
        observed = utils.port_rule_masking(port_min, port_max)
        expected = _port_rule_masking(port_min, port_max)
        self.assertCountEqual(expected, observed)

    def test_port_rule_masking_random_ranges(self):
        # calling randint a bunch of times is really slow
        randports = sorted(random.sample(range(1, 65536), 2000))
        port_max = 0
        for i in randports:
            port_min = port_max
            port_max = i
            self.compare_port_ranges_results(port_min, port_max)

    def test_port_rule_masking_edge_cases(self):
        # (port_min, port_max) tuples
        TESTING_DATA = [
            (5, 12),
            (20, 130),
            (4501, 33057),
            (0, 65535),
            (22, 22),
            (5001, 5001),
            (0, 7),
            (8, 15),
            (1, 127),
        ]
        for port_min, port_max in TESTING_DATA:
            self.compare_port_ranges_results(port_min, port_max)


class TestExcDetails(base.BaseTestCase):

    def test_attach_exc_details(self):
        e = Exception()
        utils.attach_exc_details(e, 'details')
        self.assertEqual('details', utils.extract_exc_details(e))

    def test_attach_exc_details_with_interpolation(self):
        e = Exception()
        utils.attach_exc_details(e, 'details: %s', 'foo')
        self.assertEqual('details: foo', utils.extract_exc_details(e))

    def test_attach_exc_details_with_None_interpolation(self):
        e = Exception()
        utils.attach_exc_details(e, 'details: %s', None)
        self.assertEqual(
            'details: %s' % str(None), utils.extract_exc_details(e))

    def test_attach_exc_details_with_multiple_interpolation(self):
        e = Exception()
        utils.attach_exc_details(
            e, 'details: %s, %s', ('foo', 'bar'))
        self.assertEqual('details: foo, bar', utils.extract_exc_details(e))

    def test_attach_exc_details_with_dict_interpolation(self):
        e = Exception()
        utils.attach_exc_details(
            e, 'details: %(foo)s, %(bar)s', {'foo': 'foo', 'bar': 'bar'})
        self.assertEqual('details: foo, bar', utils.extract_exc_details(e))

    def test_extract_exc_details_no_details_attached(self):
        self.assertIsInstance(
            utils.extract_exc_details(Exception()), str)


@ddt.ddt
class ImportModulesRecursivelyTestCase(base.BaseTestCase):

    @ddt.data('/', r'\\')
    def test_recursion(self, separator):
        expected_modules = (
            'neutron.tests.unit.tests.example.dir.example_module',
            'neutron.tests.unit.tests.example.dir.subdir.example_module',
        )
        for module in expected_modules:
            sys.modules.pop(module, None)

        topdir = re.sub(r'[/\\]+', separator, os.path.dirname(tests.__file__))
        modules = utils.import_modules_recursively(topdir)
        for module in expected_modules:
            self.assertIn(module, modules)
            self.assertIn(module, sys.modules)


class TestThrottler(base.BaseTestCase):
    def test_throttler(self):
        threshold = 1
        orig_function = mock.Mock()
        # Add this magic name as it's required by functools
        orig_function.__name__ = 'mock_func'
        throttled_func = utils.throttler(threshold)(orig_function)

        throttled_func()

        sleep = utils.time.sleep

        def sleep_mock(amount_to_sleep):
            sleep(amount_to_sleep)
            self.assertGreaterEqual(threshold, amount_to_sleep)

        with mock.patch.object(utils.time, "sleep",
                               side_effect=sleep_mock):
            throttled_func()

        self.assertEqual(2, orig_function.call_count)

        lock_with_timer = throttled_func.__closure__[1].cell_contents
        timestamp = lock_with_timer.timestamp - threshold
        lock_with_timer.timestamp = timestamp

        throttled_func()

        self.assertEqual(3, orig_function.call_count)
        self.assertLess(timestamp, lock_with_timer.timestamp)

    def test_method_docstring_is_preserved(self):
        class Klass:
            @utils.throttler()
            def method(self):
                """Docstring"""

        self.assertEqual("Docstring", Klass.method.__doc__)

    def test_method_still_callable(self):
        class Klass:
            @utils.throttler()
            def method(self):
                pass

        obj = Klass()
        obj.method()


class BaseUnitConversionTest:

    def test_bytes_to_bits(self):
        test_values = [
            (0, 0),  # 0 bytes should be 0 bits
            (1, 8)   # 1 byte should be 8 bits
        ]
        for input_bytes, expected_bits in test_values:
            self.assertEqual(
                expected_bits, utils.bytes_to_bits(input_bytes)
            )


class TestSIUnitConversions(BaseUnitConversionTest, base.BaseTestCase):

    base_unit = constants.SI_BASE

    def test_bits_to_kilobits(self):
        test_values = [
            (0, 0),  # 0 bites should be 0 kilobites
            (1, 1),  # 1 bit should be 1 kilobit
            (999, 1),  # 999 bits should be 1 kilobit
            (1000, 1),  # 1000 bits should be 1 kilobit
            (1001, 2)   # 1001 bits should be 2 kilobits
        ]
        for input_bits, expected_kilobits in test_values:
            self.assertEqual(
                expected_kilobits,
                utils.bits_to_kilobits(input_bits, self.base_unit)
            )


class TestIECUnitConversions(BaseUnitConversionTest, base.BaseTestCase):

    base_unit = constants.IEC_BASE

    def test_bits_to_kilobits(self):
        test_values = [
            (0, 0),  # 0 bites should be 0 kilobites
            (1, 1),  # 1 bit should be 1 kilobit
            (1023, 1),  # 1023 bits should be 1 kilobit
            (1024, 1),  # 1024 bits should be 1 kilobit
            (1025, 2)   # 1025 bits should be 2 kilobits
        ]
        for input_bits, expected_kilobits in test_values:
            self.assertEqual(
                expected_kilobits,
                utils.bits_to_kilobits(input_bits, self.base_unit)
            )


class TestRpBandwidthValidator(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        self.device_name_set = {'ens4', 'ens7'}
        self.valid_rp_bandwidths = {
            'ens7': {'egress': 10000, 'ingress': 10000}
        }
        self.not_valid_rp_bandwidth = {
            'ens8': {'egress': 10000, 'ingress': 10000}
        }

    def test_validate_rp_bandwidth_with_device_names(self):
        try:
            utils.validate_rp_bandwidth(self.valid_rp_bandwidths,
                                        self.device_name_set)
        except ValueError:
            self.fail("validate_rp_bandwidth failed to validate %s" %
                      self.valid_rp_bandwidths)

        self.assertRaises(ValueError, utils.validate_rp_bandwidth,
                          self.not_valid_rp_bandwidth, self.device_name_set)


class SpawnWithOrWithoutProfilerTestCase(base.BaseTestCase):

    def _compare_profilers_in_parent_and_in_child(self, init_profiler):

        q = queue.Queue()

        def is_profiler_initialized(where):
            # Instead of returning a single boolean add information so we can
            # identify which thread produced the result without depending on
            # queue order.
            return {where: bool(profiler.get())}

        def thread_with_no_leaked_profiler():
            if init_profiler:
                profiler.init(hmac_key='fake secret')

            utils.spawn_n(lambda: q.put(is_profiler_initialized('in-child')))
            q.put(is_profiler_initialized('in-parent'))

        # Make sure in parent we start with an uninitialized profiler by
        # spawning a new thread. Otherwise the unit test runner thread may
        # leak an initialized profiler from one test to another.
        _thread = threading.Thread(target=thread_with_no_leaked_profiler)
        _thread.start()

        # In order to have some global protection against leaking initialized
        # profilers neutron.test.base.BaseTestCase.setup() also calls
        # addCleanup(profiler.clean)

        # Merge the results independently of queue order.
        results = {}
        results.update(q.get())
        results.update(q.get())

        self.assertEqual(
            {'in-parent': init_profiler,
             'in-child': init_profiler},
            results)

    def test_spawn_with_profiler(self):
        self._compare_profilers_in_parent_and_in_child(init_profiler=True)

    def test_spawn_without_profiler(self):
        self._compare_profilers_in_parent_and_in_child(init_profiler=False)


@utils.SingletonDecorator
class _TestSingletonClass1:
    def __init__(self, variable):
        self.variable = variable


@utils.SingletonDecorator
class _TestSingletonClass2:
    def __init__(self, variable):
        self.variable = variable


class SingletonDecoratorTestCase(base.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.instance_1 = self.instance_2 = None
        self.addCleanup(self._destroy_instances)

    def _destroy_instances(self):
        try:
            del self.instance_1
        except AttributeError:
            pass
        try:
            del self.instance_2
        except AttributeError:
            pass

    def test_singleton_instance_class(self):
        self.instance_1 = _TestSingletonClass1('value1')
        self.instance_2 = _TestSingletonClass1('other_value_that_is_not_set')
        self.assertIs(self.instance_1, self.instance_2)
        self.assertEqual('value1', self.instance_2.variable)

        # If one reference is deleted, the other variable (instance_2) is still
        # referenced to the same object and the singleton decorator keeps it.
        del self.instance_1
        self.instance_1 = _TestSingletonClass1('again_value_that_is_not_set')
        self.assertIs(self.instance_1, self.instance_2)
        self.assertEqual('value1', self.instance_2.variable)

    def test_singleton_several_independent_classes(self):
        self.instance_1 = _TestSingletonClass1('value1')
        self.instance_2 = _TestSingletonClass2('value2')
        self.assertIsNot(self.instance_1, self.instance_2)
        self.assertEqual('value1', self.instance_1.variable)
        self.assertEqual('value2', self.instance_2.variable)

    def test_singleton_instance_deletion(self):
        self.instance_1 = _TestSingletonClass1('value1')
        self.instance_2 = _TestSingletonClass2('value2')
        self.assertIsNotNone(self.instance_1)
        self.assertIsNotNone(self.instance_2)

        # Delete instance 1, new value "value1_b" set in __init__()
        del self.instance_1
        self.instance_1 = _TestSingletonClass1('value1_b')
        self.instance_2 = _TestSingletonClass2('value2_b')
        self.assertIsNotNone(self.instance_1)
        self.assertIsNotNone(self.instance_2)
        self.assertEqual('value1_b', self.instance_1.variable)
        self.assertEqual('value2', self.instance_2.variable)

        # Delete instance 2, new value "value2_c" set in __init__()
        del self.instance_2
        self.instance_1 = _TestSingletonClass1('value1_c')
        self.instance_2 = _TestSingletonClass2('value2_c')
        self.assertIsNotNone(self.instance_1)
        self.assertIsNotNone(self.instance_2)
        self.assertEqual('value1_b', self.instance_1.variable)
        self.assertEqual('value2_c', self.instance_2.variable)


class DisableNotificationTestCase(base.BaseTestCase):

    @utils.disable_notifications
    def sample_method(self):
        raise AttributeError()

    def test_notification_rpc_workers_lt_one(self):
        cfg.CONF.set_override('rpc_workers', 0)
        self.assertIsNone(self.sample_method())

    def test_notification_rpc_workers_none(self):
        cfg.CONF.set_override('rpc_workers', None)
        self.assertRaises(AttributeError, self.sample_method)

    def test_notification_rpc_workers_one(self):
        cfg.CONF.set_override('rpc_workers', 1)
        self.assertRaises(AttributeError, self.sample_method)


class SignatureTestCase(base.BaseTestCase):

    def test_sign_instance_id(self):
        conf = mock.Mock()
        conf.metadata_proxy_shared_secret = 'secret'
        self.assertEqual(
            '773ba44693c7553d6ee20f61ea5d2757a9a4f4a44d2841ae4e95b52e4cd62db4',
            utils.sign_instance_id(conf, 'foo')
        )


class ParsePermittedEthertypesTestCase(base.BaseTestCase):

    @mock.patch.object(utils, 'LOG')
    def test_parse_permitted_ethertypes(self, mock_log):
        permitted_ethertypes = ['0x1234',
                                '0x1234',
                                '0x5678',
                                '0XCAfe',
                                '1112',
                                '0x123R']
        ret = utils.parse_permitted_ethertypes(permitted_ethertypes)
        self.assertEqual({'0x1234', '0x5678', '0xcafe'}, ret)
        calls = [mock.call('Custom ethertype %s is repeated', '0x1234'),
                 mock.call('Custom ethertype %s is not a hexadecimal number.',
                           '1112'),
                 mock.call('Custom ethertype %s is not a hexadecimal number.',
                           '0x123R'),
                 ]
        mock_log.warning.assert_has_calls(calls)


class StringMapTestCase(base.BaseTestCase):
    data = {'a': 1, 'b': None, 'c': 'hi'}

    def test_stringmap(self):
        expected = {'a': '1', 'b': '', 'c': 'hi'}

        self.assertEqual(expected, utils.stringmap(self.data))

    def test_stringmap_custom_default(self):
        self.assertEqual('None', utils.stringmap(self.data, 'None')['b'])


class ThreadPoolExecutorWithBlockTestCase(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        self._queue = queue.Queue()

    def _check_values(self, expected_values):
        # Retrieve all stored values
        stored_values = []
        while not self._queue.empty():
            stored_values.append(self._queue.get())

        for expected_value in expected_values:
            self.assertIn(expected_value, stored_values)

    def _add_tasks(self, tpool, num_tasks):
        expected_values = []
        for idx in range(num_tasks):
            value1 = 'value1_%s' % idx
            value2 = 'value2_%s' % idx
            values = (value1, value2)
            expected_values.append(values)
            tpool.submit(self._task, self._queue, value1, value2=value2)
        return expected_values

    @staticmethod
    def _task(_queue, value1, value2=None):
        _queue.put((value1, value2))
        time.sleep(1)

    def test_simple_execution(self):
        max_workers = 5
        t1 = timeutils.utcnow()
        with utils.ThreadPoolExecutorWithBlock(max_workers=max_workers) as \
                tpool:
            expected_values = self._add_tasks(tpool, max_workers * 3)

        utils.wait_until_true(lambda: self._queue.qsize() == max_workers * 3,
                              timeout=5, sleep=0.1)
        t2 = timeutils.utcnow()
        diff_seconds = (t2 - t1).total_seconds()
        # NOTE(ralonsoh): we can't expect a 3 seconds sharp time gap, but at
        # least the execution is contained in this interval. If each task takes
        # 1 second, 5 threads can be executed in parallel and 15 workers are
        # required, that will take at least 3 seconds and should complete in
        # 6 seconds (2x).
        self.assertTrue(3 < diff_seconds < 6)

        self._check_values(expected_values)

    def test_simple_execution_fast_exit(self):
        max_workers = 5
        t1 = timeutils.utcnow()
        with utils.ThreadPoolExecutorWithBlock(max_workers=max_workers) as \
                tpool:
            expected_values = self._add_tasks(tpool, max_workers)
            tpool.shutdown(wait=False)
            self._add_tasks(tpool, max_workers)

        utils.wait_until_true(lambda: self._queue.qsize() == max_workers,
                              timeout=5, sleep=0.1)
        t2 = timeutils.utcnow()
        diff_seconds = (t2 - t1).total_seconds()
        # Similar to above, execution will take at least 1 second and should
        # complete in 2 seconds (2x).
        self.assertTrue(1 < diff_seconds < 2)
        self._check_values(expected_values)


class TestFindUniqueSequence(base.BaseTestCase):
    def test_empty_groups(self):
        # Empty input should return empty list
        result = utils.find_unique_sequence([])
        self.assertEqual([], result)

    def test_single_group_single_value(self):
        # Single group with one value
        result = utils.find_unique_sequence([['a']])
        self.assertEqual(['a'], result)

    def test_single_group_multiple_values(self):
        # Single group with multiple values - returns first value
        result = utils.find_unique_sequence([['a', 'b', 'c']])
        self.assertEqual(['a'], result)

    def test_multiple_groups_no_overlap(self):
        # Multiple groups with distinct values - straightforward case
        result = utils.find_unique_sequence([['a'], ['b'], ['c']])
        self.assertEqual(['a', 'b', 'c'], result)

    def test_multiple_groups_with_overlap_solvable(self):
        # Groups with overlapping values but a solution exists
        result = utils.find_unique_sequence(
            [['a', 'b'], ['b', 'c'], ['c', 'a']])
        self.assertEqual(['a', 'b', 'c'], result)

    def test_backtracking_required(self):
        # First choice in group 1 would conflict, requiring backtrack
        # Group 0 picks 'a', group 1 must pick 'b' (not 'a'),
        # group 2 must pick 'c'
        result = utils.find_unique_sequence([['a'], ['a', 'b'], ['b', 'c']])
        self.assertEqual(['a', 'b', 'c'], result)

    def test_no_solution_all_same_value(self):
        # All groups have only the same value - no valid sequence
        result = utils.find_unique_sequence([['a'], ['a'], ['a']])
        self.assertIsNone(result)

    def test_no_solution_insufficient_unique_values(self):
        # Not enough unique values across groups
        result = utils.find_unique_sequence(
            [['a', 'b'], ['a', 'b'], ['a', 'b']])
        self.assertIsNone(result)

    def test_complex_backtracking(self):
        # Requires multiple levels of backtracking
        # If we pick 'x' first, then 'y', we're stuck at group 2
        # Need to backtrack and try different combinations
        groups = [['x', 'a'], ['y', 'x', 'b'], ['x', 'y', 'c']]
        result = utils.find_unique_sequence(groups)
        # Verify result is valid: one per group, all unique
        self.assertIsNotNone(result)
        self.assertEqual(3, len(result))
        self.assertEqual(3, len(set(result)))
        for i, val in enumerate(result):
            self.assertIn(val, groups[i])

    def test_result_preserves_group_order(self):
        # Result should have one element from each group in order
        groups = [['1', '2'], ['3', '4'], ['5', '6']]
        result = utils.find_unique_sequence(groups)
        self.assertEqual(3, len(result))
        self.assertIn(result[0], groups[0])
        self.assertIn(result[1], groups[1])
        self.assertIn(result[2], groups[2])

    def test_partial_overlap(self):
        # Some groups share values, but solution exists
        groups = [['a', 'b', 'c'], ['b', 'c', 'd'], ['c', 'd', 'e']]
        result = utils.find_unique_sequence(groups)
        self.assertIsNotNone(result)
        self.assertEqual(3, len(result))
        self.assertEqual(3, len(set(result)))

    def test_single_empty_group_in_list(self):
        # One of the groups is empty - no solution possible
        result = utils.find_unique_sequence([['a'], [], ['c']])
        self.assertIsNone(result)
